#ifndef __P3PMatchInfo_H__
#define __P3PMatchInfo_H__

#include "GURL.h"
#include "GContainer.h"
#include "GString.h"
#include "P3PEvaluator.h"


class CP3PObject;
class P3PMatchInfo;

class CookieInfo:public GPEnabled
{
	friend CP3PObject;
	friend P3PMatchInfo;
	GString name;
	GString value;
	GString domain;
	GString path;
	bool blocked;
	bool needpolicycheck;
	bool thirdparty;

	CookieInfo():blocked(false), needpolicycheck(true), thirdparty(false){}
	CookieInfo(const char *_name, const char *_value, const char *_domain, const char *_path):
			blocked(false), needpolicycheck(false), thirdparty(false)
			{
				name=_name;
				value=_value;
				domain=_domain;
				path=_path;
			}

	
};

class P3PMatchInfo
{
	
	class URLTypeInfo
	{
		friend P3PMatchInfo;
		friend CP3PObject;
		GString type;
		GURL URL;
	public:
		URLTypeInfo(const GString &t="", const GURL url=""):type(t), URL(url) {}
	};

	GURL mainURL;
	
	Result result;
	
	GMap<GURL, URLTypeInfo> embededURLs;
	GMap<GURL, Result> policyURLs;
	GCriticalSection InfoLock;
	GPList<CookieInfo> cookielist;
	GCriticalSection cookielock;
	bool valid_object;

public:
	enum P3PMatch{ UNKNOWN=-1, MATCH, UNMATCH, MATCH_NOTHIRDPARTY, UNKNOWN_MATCHTHIRDPARTY };

	void AddEmbededURL(const GURL &url, const GString &type);
	void Clean();
	void CleanInfo();
	void CleanCookies();
	P3PMatch GetNetResult(bool &blocked_cookies);
	void CheckEmbeds();
	void SetMainURL(GURL &url) { mainURL = url; }
	GURL GetMainURL() { return mainURL; }
	int CheckMainURL(const GURL &url, GURL &policyURL=GURL());
	int CheckMainURLWithP3PHeader(const GURL &siteurl);
	Result GetMainResult() { return result; }
	bool P3PMatchInfo::GetEmbeddedResult(const GURL &embedURL, Result &result);
	void GetEmbeddedURLList(GList<GURL> &urllist);
	Result::P3PMatch GetEmbeddedResult(const GURL &url);
	GString GetEmbeddedURLType(const GURL &url);
	GURL GetPolicyURL(const GURL &embedURL);
	GURL GetAppliedMainPolicyURL();
	bool AddCookie(const char *name, const char *value, const char *domain, const char *trackerdomain, const char *path, const GURL &mainurl);
	bool NeedCookieBlocking(const char *name, const char *value, const char *domain, const char *trackerdomain, const char *path, const GURL &mainurl);
	void CheckAllCookiePolicies();
	void GetCookieInfo(int &cookiesdetected, int &cookiesblocked, int &othercookiesdetected, int &othercookiesblocked);
	void SetValid(bool status) { valid_object = status; }



	~P3PMatchInfo();

private:
	void EmbededURLCheck(const GURL &url);
	void CheckValid();
	

	friend void EmbededURLStart(void *obj);
	static int CheckCookiePolicy(const GString &name, const GString &value, const GString &domain, const GString &path, bool downloadifneeded=false);
	

public:
	static void PlaySoundPolicyUnkown();
	static void PlaySoundPolicyUnmatched();
	static void PlaySoundPolicyMatched();

};



#endif