// P3PObject.h : Declaration of the CP3PObject

#ifndef __P3PObject_H_
#define __P3PObject_H_

#include "resource.h"       // main symbols
#include <ExDisp.h>

#include "IconDialog.h"
#include <shlwapi.h>
#include <msHtml.h>
#include "GURL.h"
#include "GString.h"
#include "P3PMatchInfo.h"

const TCHAR REG_KEY[] = _T("Software\\MSDN\\BHO");
const TCHAR REG_VAL[] = _T("ShowWindowAtStartup");
const int H_DLGBUTTONBAR = 20;
const int W_BUTTON = 90;
const int PADDING = 5;

struct PolicyCheckData;
/////////////////////////////////////////////////////////////////////////////
// CP3PObject
class ATL_NO_VTABLE CP3PObject : 
	public CComObjectRootEx<CComSingleThreadModel>,
	public CComCoClass<CP3PObject, &CLSID_P3PObject>,
	public IObjectWithSiteImpl<CP3PObject>,
	public IDispatchImpl<IP3PObject, &IID_IP3PObject, &LIBID_P3PClientLib>
{
public:
	CP3PObject();
	/*: icondlg(this), hook(0)
	{
		m_bDocumentCompleted = false;

		RECT rc;
		SystemParametersInfo(SPI_GETWORKAREA, 0, &rc, 0);
		
		CopyRect(&m_rcIE, &rc);
		m_rcIE.bottom = (rc.bottom*6/10); 

		CopyRect(&m_rcHTML, &rc);
		m_rcHTML.top = m_rcIE.bottom; 
	}*/

	~CP3PObject();
	

DECLARE_REGISTRY_RESOURCEID(IDR_P3PObject)

DECLARE_PROTECT_FINAL_CONSTRUCT()

BEGIN_COM_MAP(CP3PObject)
	COM_INTERFACE_ENTRY(IP3PObject)
	COM_INTERFACE_ENTRY(IDispatch)
	COM_INTERFACE_ENTRY(IObjectWithSite)
END_COM_MAP()

// IObjectWithSite
public:
	STDMETHOD(SetSite)(IUnknown *pUnkSite);

// IDispatch
public:    
	STDMETHOD(Invoke)(DISPID, REFIID, LCID, WORD, DISPPARAMS*, 
		              VARIANT*, EXCEPINFO*, UINT*);

// IP3PObject	
public:
	STDMETHOD(OnDocumentComplete)(IDispatch *tDisp, BSTR *url);
	STDMETHOD(OnQuit)(void);
	STDMETHOD(GetCodeWindowRect)(LPRECT);
	void RepositionIcon();

private:
	STDMETHOD(Connect)(void);
	STDMETHOD(Disconnect)(void);
	STDMETHOD(RetrieveBrowserWindow)(void);
	BOOL DisplayP3PIcon();
	static BOOL CALLBACK WndEnumProc(HWND, LPARAM); 

private:
	CComQIPtr<IWebBrowser2, &IID_IWebBrowser2> m_spWebBrowser2;
    CComQIPtr<IConnectionPointContainer, &IID_IConnectionPointContainer> m_spCPC;
	CComPtr<IHTMLElement> m_pBody;
	BOOL m_bDocumentCompleted;
	DWORD m_dwCookie;
	static bool pageloading;
	
	RECT m_rcIE, m_rcHTML;
public:
	CIconDialog icondlg;


private:
	void OnBeforeNavigate(IDispatch *pDisp,
    GString &URL,
    DWORD &Flags,
    GString &TargetFrameName,
    GString &PostData,
    GString &Headers,
    BOOL &Cancel);
public:
	HHOOK hook;
	WNDPROC wndproc;
	HWND m_hwndBrowser;

public:
	int PolicyCheck(GURL &URL,DWORD &Flags,GString &TargetFrameName,GString &PostData,GString &Headers);
	GURL get_locationURL()
	{
		return GetLocationURL();
	}
	P3PMatchInfo &GetMatchInfo()
	{
		return matchinfo;
	}
	static CP3PObject *GetP3PObjectForThread(DWORD threadID);
	static bool AddP3PObjectForThread(DWORD threadID, CP3PObject *p3p);
	static bool DeleteP3PObjectForThread(DWORD threadID);
	static int  GetP3PObjectMapSize();

	GURL locationURL;
	///returns blocking status, if true block the cookie
	bool AddCookie(const char *name, const char *value, const char *domain, const char *trackerdomain, const char *path);
	bool NeedCookieBlocking(const char *name, const char *value, const char *domain, const char *trackerdomain, const char *path);
	void InformRedirection(const char *OriginURL, const char *RedirectedURL);
	void EvaluatePolicy(PolicyCheckData* data = 0, bool waitforthread=true);

private:
	int PostCheck(GURL &URL, Result &result);
	GURL GetLocationURL();
	GURL FindLinkReferenceInDocument(CComQIPtr<IWebBrowser2, &IID_IWebBrowser2> m_spWebBrowser2);
	void GetAllEmbeddedURLs(CComQIPtr<IHTMLDocument2, &IID_IHTMLDocument2> htmldoc, GList<P3PMatchInfo::URLTypeInfo> &list);
	void GetAllEmbeddedURLs(IWebBrowser2 &browser, GList<P3PMatchInfo::URLTypeInfo> &list);
	bool IsTopFrame(IDispatch *pDisp);


	P3PMatchInfo matchinfo;

	static GMap<DWORD, CP3PObject*> p3pobjectmap;
	static GCriticalSection p3pobjectlock;
	static GMap<DWORD, GMonitor*> workerthreadmonitormap;

	static GCriticalSection workerthreadmonitormaplock;
	GMonitor workerthreadmonitor;
	GThread *workerthread;
	GList<P3PMatchInfo::URLTypeInfo> embed_list;
	GURL linkrefurl;
	bool isPostRequest;

	
public:
	
	static GMonitor* GetWorkerThreadMonitor(DWORD threadID)
	{ 
		GCriticalSectionLock lock(&workerthreadmonitormaplock);
		if( GPosition pos = workerthreadmonitormap)
			return (workerthreadmonitormap[pos]); 
		else
			return NULL;
	}
	void AddWorkerThreadMonitor(DWORD threadID)
	{
		GCriticalSectionLock lock(&workerthreadmonitormaplock);
		if( workerthreadmonitormap.contains(threadID) )
			return;
		workerthreadmonitormap[threadID] = &workerthreadmonitor;
	}
	void DeleteWorkerThreadMonitor(DWORD threadID)
	{
		GCriticalSectionLock lock(&workerthreadmonitormaplock);
		if( GPosition pos = workerthreadmonitormap.contains(threadID) )
			workerthreadmonitormap.del(pos);
	}

};

#include "CookiemanCalls.h"

extern TCM_RegisterBho *CM_RegisterBho;
extern TCM_UnregisterBho *CM_UnregisterBho;
extern TCM_UpdateTrackersBlackList *CM_UpdateTrackersBlackList;
extern TCM_EnableIDcide *CM_EnableIDcide;
extern TCM_CleanRequests *CM_CleanRequests;
extern TCM_GetProxySettings *CM_GetProxySettings;



#endif //__P3PObject_H_
