// P3PObject.cpp : Implementation of CP3PObject
#include "stdafx.h"
#include "P3PClient.h"
#include "P3PObject.h"
#include "ReadHTTPFile.h"
#include <ExDispID.h>
#include <stdio.h>
#include <comdef.h>
#include "XmlNode.h"
#include "PolicyHandler.h"
#include "Globals.h"
#include "P3PEvaluator.h"
#include "P3PPreferences.h"
#include "PostWarningDialog.h"


#define THREAD_WAIT_TIME 100   ///milliseconds

GMap<DWORD, CP3PObject*> CP3PObject::p3pobjectmap;
GCriticalSection CP3PObject::p3pobjectlock;

TCM_RegisterBho *CM_RegisterBho=0;
TCM_UnregisterBho *CM_UnregisterBho=0;
TCM_UpdateTrackersBlackList *CM_UpdateTrackersBlackList=0;
TCM_EnableIDcide *CM_EnableIDcide=0;
TCM_CleanRequests *CM_CleanRequests=0;
TCM_GetProxySettings *CM_GetProxySettings=0;

GMap<DWORD, GMonitor*> CP3PObject::workerthreadmonitormap;

GCriticalSection CP3PObject::workerthreadmonitormaplock;


struct PolicyCheckData
{
	GURL url;
	DWORD flags;
	GString TargetFrameName;
	GString PostData;
	GString Headers;
	CP3PObject *object;
};
void PolicyCheckStart(void *arg);



/////////////////////////////////////////////////////////////////////////////
// Global variables used in this module
//static GMap<DWORD, HHOOK> hookmap;
//static HHOOK g_hWndHook = NULL;
static HHOOK g_hKbHook = NULL;
/////////////////////////////////////////////////////////////////////////////
// CP3PObject

LRESULT CALLBACK WindowProc(
  HWND hwnd,      // handle to window
  UINT uMsg,      // message identifier
  WPARAM wParam,  // first message parameter
  LPARAM lParam   // second message parameter
)
{
	CP3PObject *vs = (CP3PObject*)GetProp(hwnd, _T("p3pinstance"));
			
	if( vs )
	{
		switch(uMsg)
		{
		case WM_KEYDOWN:
			if( wParam == VK_F12 )
				vs->icondlg.PostMessage(WM_LBUTTONDOWN, 0, 0);
			break;
		case WM_ENTERSIZEMOVE:
		case WM_EXITSIZEMOVE:
		case WM_SIZING:
		case WM_SIZE:
		case WM_MOVE:
		case WM_MOVING:
			vs->RepositionIcon();

		}
		return CallWindowProc(vs->wndproc, hwnd, uMsg, wParam, lParam);
	}
	return 0;
}


LRESULT CALLBACK IEWndProc(int nCode, 
	WPARAM wParam, LPARAM lParam)
{
	// Typical start-off for any hook
	CP3PObject *p3p = 0;//CP3PObject::GetP3PObjectForThread(GetCurrentThreadId());
	if( CP3PObject::GetP3PObjectMapSize() > 0 )
		p3p = CP3PObject::GetP3PObjectForThread(GetCurrentThreadId());
	if( nCode <0 )
		return CallNextHookEx(p3p?p3p->hook:NULL, nCode, wParam, lParam);

	CWPSTRUCT *cwp = (CWPSTRUCT*)lParam;

	if( nCode == HC_ACTION && cwp->hwnd)
	{
		CP3PObject *vs = (CP3PObject*)GetProp(cwp->hwnd, _T("p3pinstance"));
		if( vs && p3p && vs == p3p)
		{
			switch(cwp->message)
			{
			case WM_ENTERSIZEMOVE:
			case WM_EXITSIZEMOVE:
			case WM_SIZING:
			case WM_SIZE:
			case WM_MOVE:
			case WM_MOVING:
				p3p->RepositionIcon();
				break;
			}
		}
	}

	// Typical end for any hook procedure
	return CallNextHookEx(p3p?p3p->hook:NULL, nCode, wParam, lParam);
}



/////////////////////////////////////////////////////////////////////////////
// IEKeyboardProc - Keyboard hook procedure. It's a static member
// of the class and needs g_pThis to access the class members
// 
LRESULT CALLBACK IEKeyboardProc(int nCode, 
	WPARAM wParam, LPARAM lParam)
{
	// Typical start-off for any hook
	if( nCode <0 )
		return CallNextHookEx(g_hKbHook, nCode, wParam, lParam);

	// Process the key only once
	if ((lParam & 0x80000000) || (lParam & 0x40000000))
		return CallNextHookEx(g_hKbHook, nCode, wParam, lParam);

	// The key we're interested in is F12. If it's been pressed
	// then forces the Code Window to be refreshed
	if (wParam==VK_F12)	
	{
		CP3PObject *p3p = CP3PObject::GetP3PObjectForThread(GetCurrentThreadId());
		if( p3p )
			p3p->icondlg.PostMessage(WM_LBUTTONDOWN);
	}

	// Typical end for any hook procedure
	return CallNextHookEx(g_hKbHook, nCode, wParam, lParam);
}

//////Constructor and destructor
CP3PObject::CP3PObject() : icondlg(this), wndproc(0)
{
	m_bDocumentCompleted = false;
	isPostRequest = false;
	workerthread = 0;

	RECT rc;
	SystemParametersInfo(SPI_GETWORKAREA, 0, &rc, 0);
	
	CopyRect(&m_rcIE, &rc);
	m_rcIE.bottom = (rc.bottom*6/10); 

	CopyRect(&m_rcHTML, &rc);
	m_rcHTML.top = m_rcIE.bottom; 

	if( GetP3PObjectMapSize() == 0 )
		PolicyReferenceHandler::LoadPolicyCache();

	AddP3PObjectForThread(GetCurrentThreadId(), this);

}

CP3PObject::~CP3PObject()
{
	if( !DeleteP3PObjectForThread(GetCurrentThreadId()) )
		MessageBox(this->m_hwndBrowser, "P3P object for thread not found", "Error", MB_OK|MB_ICONSTOP);

	if( GetP3PObjectMapSize() == 0)
	{
		g_p3p_preferences.SaveCookieHistory();
		PolicyReferenceHandler::SavePolicyCache();
	}

	////////////////////////////////////////////
	//icondlg.DestroyWindow();
	if( workerthread )
	{
		while(workerthread->wait(THREAD_WAIT_TIME))
		{
			workerthreadmonitor.enter();
			workerthreadmonitor.broadcast();
			workerthreadmonitor.leave();
		}
		delete workerthread;
		workerthread = 0;
	}
	
}

CP3PObject* CP3PObject::GetP3PObjectForThread(DWORD threadID)
{
	GCriticalSectionLock lock(&p3pobjectlock);
	if( GPosition pos = p3pobjectmap.contains(threadID) )
	{
		return p3pobjectmap[pos];
	}
	return NULL;
}

bool CP3PObject::AddP3PObjectForThread(DWORD threadID, CP3PObject *p3p)
{
	GCriticalSectionLock lock(&p3pobjectlock);
	if( threadID && p3p )
	{
		p3pobjectmap[threadID] = p3p;
		return true;
	}
	return false;
}

bool CP3PObject::DeleteP3PObjectForThread(DWORD threadID)
{
	GCriticalSectionLock lock(&p3pobjectlock);
	if( GPosition pos = p3pobjectmap.contains(threadID) )
	{
		p3pobjectmap.del(pos);
		return true;
	}
	return false;
}

int  CP3PObject::GetP3PObjectMapSize()
{
	GCriticalSectionLock lock(&p3pobjectlock);
	return p3pobjectmap.size();
}

/////////////////////////////////////////////////////////////////////////////
// SetSite - Initializes the BHO
// 
HRESULT CP3PObject::SetSite(IUnknown *pUnkSite)
{
	// Retrieve and store the IWebBrowser2 pointer 
	m_spWebBrowser2 = pUnkSite; 
	if (m_spWebBrowser2 == NULL)
		return E_INVALIDARG;
	
	// Retrieve and store the IConnectionPointerContainer pointer
	m_spCPC = m_spWebBrowser2;
	if (m_spCPC == NULL) 
		return E_POINTER;

	// Retrieve and store the HWND of the browser
	RetrieveBrowserWindow();

	SetProp(m_hwndBrowser, _T("p3pinstance"), (HANDLE)this);
	
	OSVERSIONINFO osvi = {sizeof(OSVERSIONINFO)};
	GetVersionEx (&osvi);
	bool bIsWindows98orLater = (osvi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS) &&
								( (osvi.dwMajorVersion > 4) || ( (osvi.dwMajorVersion == 4) && (osvi.dwMinorVersion > 0) ) );
 
	if( bIsWindows98orLater )
		wndproc = (WNDPROC)::SetWindowLong(m_hwndBrowser, GWL_WNDPROC, (LONG)(WNDPROC)WindowProc);
	else
	{
		//g_hWndHook = SetWindowsHookEx(WH_CALLWNDPROC, IEWndProc, 0, 0/*GetCurrentThreadId()*/);
		//g_hKbHook = SetWindowsHookEx( WH_KEYBOARD, IEKeyboardProc, NULL, GetCurrentThreadId() );
		hook = SetWindowsHookEx(WH_CALLWNDPROC, IEWndProc, 0, GetCurrentThreadId());
	}
	
	DisplayP3PIcon();

	// Connect to the container for receiving event notifications
	return Connect();
}


/////////////////////////////////////////////////////////////////////////////
// Connect - Register to the container as an event handler
// 
HRESULT CP3PObject::Connect(void)
{
	HRESULT hr;
	CComPtr<IConnectionPoint> spCP;

	// Receives the connection point for WebBrowser events
    hr = m_spCPC->FindConnectionPoint(
			DIID_DWebBrowserEvents2, 
			&spCP);
    if (FAILED(hr))
		return hr;

    // Pass the event handlers to the container
	hr = spCP->Advise( 
			reinterpret_cast<IDispatch*>(this), 
			&m_dwCookie);

    return hr; 
}


/////////////////////////////////////////////////////////////////////////////
// Disconnect - Unregister as an event handler module
// 
HRESULT CP3PObject::Disconnect(void)
{
	HRESULT hr;
	CComPtr<IConnectionPoint> spCP;

	// Receives the connection point for WebBrowser events
    hr = m_spCPC->FindConnectionPoint(
			DIID_DWebBrowserEvents2, 
			&spCP);
    if (FAILED(hr))
		return hr;

    // Stop getting event notifications
	hr = spCP->Unadvise(m_dwCookie);
    return hr;
}


/////////////////////////////////////////////////////////////////////////////
// Invoke - Switch-procedure for events
// 
HRESULT CP3PObject::Invoke(DISPID dispidMember, REFIID riid, LCID lcid, WORD wFlags,
	DISPPARAMS* pDispParams, VARIANT* pvarResult,
    EXCEPINFO*  pExcepInfo,  UINT* puArgErr)
{
   if (!pDispParams)
      return E_INVALIDARG;

   if( !g_enabled )
   {
	   if( dispidMember == DISPID_DOCUMENTCOMPLETE )
	   {
			m_bDocumentCompleted = true;
	   }
	   else if( dispidMember == DISPID_BEFORENAVIGATE2 )
	   {
			m_bDocumentCompleted = false;
	   }
	   return S_OK;
   }
 
   if (dispidMember == DISPID_BEFORENAVIGATE2)
   {
		VARIANTARG varg0;

		// Make sure the wFlags are valid.
		if(wFlags & ~(DISPATCH_METHOD | DISPATCH_PROPERTYGET |
			DISPATCH_PROPERTYPUT | DISPATCH_PROPERTYPUTREF))
			return E_INVALIDARG;

		// This object only exposes a "default" interface.
		if(!IsEqualIID(riid, IID_NULL))
			return ResultFromScode(DISP_E_UNKNOWNINTERFACE);

		VariantInit(&varg0);
		UINT puArgErr=0;

		IDispatch *pDisp;
		GString URL,  TargetFrameName, Headers;
		GString  PostData;
		DWORD  Flags=0;
		BOOL Cancel=FALSE;

		DispGetParam(pDispParams, 0, VT_DISPATCH, &varg0, &puArgErr);
		pDisp = varg0.pdispVal;

		DispGetParam(pDispParams, 1, VT_BSTR, &varg0, &puArgErr);
		URL = cvt_ws_to_utf8(varg0.bstrVal);


		DispGetParam(pDispParams, 3, VT_BSTR, &varg0, &puArgErr);
		TargetFrameName = cvt_ws_to_utf8(varg0.bstrVal);

	
		VARIANT* vtPostedData = V_VARIANTREF(&pDispParams->rgvarg[2]);
		vtPostedData = V_VARIANTREF(vtPostedData);
		if (V_VT(vtPostedData) & VT_ARRAY)
		{
			// must be a vector of bytes
			
			if(!(vtPostedData->parray->cDims == 1 && vtPostedData->parray->cbElements == 1))
				MessageBox(this->m_hwndBrowser, NULL, "Error", MB_OK);

			vtPostedData->vt |= VT_UI1;


			SAFEARRAY *psa = (SAFEARRAY *)vtPostedData->parray;

			char *str;
			HRESULT hr = SafeArrayAccessData(psa, (void**)&str);
			if( SUCCEEDED(hr) )
			{
				PostData = str;
				SafeArrayUnaccessData(psa);
			}
			
		}


		DispGetParam(pDispParams, 5, VT_BSTR, &varg0, &puArgErr);
		Headers = cvt_ws_to_utf8(varg0.bstrVal);

	    OnBeforeNavigate(pDisp, URL, Flags, TargetFrameName, PostData, Headers, Cancel);
		if( Cancel == TRUE )
			*(pDispParams->rgvarg[0].pboolVal) = VARIANT_TRUE;

   }

   // Document complete
   if (dispidMember == DISPID_DOCUMENTCOMPLETE /*|| dispidMember == DISPID_NAVIGATECOMPLETE2*/)
   {
		VARIANTARG varg0;
		// Make sure the wFlags are valid.
		if(wFlags & ~(DISPATCH_METHOD | DISPATCH_PROPERTYGET |
			DISPATCH_PROPERTYPUT | DISPATCH_PROPERTYPUTREF))
			return E_INVALIDARG;

		// This object only exposes a "default" interface.
		if(!IsEqualIID(riid, IID_NULL))
			return ResultFromScode(DISP_E_UNKNOWNINTERFACE);

		VariantInit(&varg0);

		IDispatch *tDisp;
		BSTR tURL;

		DispGetParam(pDispParams, 0, VT_DISPATCH, &varg0, puArgErr);
		tDisp = varg0.pdispVal;

		DispGetParam(pDispParams, 1, VT_BSTR, &varg0, puArgErr);
		tURL = varg0.bstrVal;

		OnDocumentComplete(tDisp, &tURL);

   }

   // Download complete (to catch refresh and F5)
   if (dispidMember == DISPID_DOWNLOADCOMPLETE)
   {
	   if (m_bDocumentCompleted)
		OnDocumentComplete(NULL, NULL);
		
   }

   // Quit
   if (dispidMember == DISPID_ONQUIT)
		OnQuit();

   return S_OK;
}


/////////////////////////////////////////////////////////////////////////////
// OnDocumentComplete - A new document's been initialized
// and is available through the WebBrowser's Document property

// 
HRESULT CP3PObject::OnDocumentComplete(IDispatch *pDisp, BSTR *url)
{
	///exclude all other frames except main
	if( pDisp )
	{
		GURL frameurl;
		if( url )
			frameurl = cvt_ws_to_utf8(*url);

		if( IsTopFrame(pDisp) )
		{
			m_bDocumentCompleted = true;

			PolicyCheckData* pcdata = new PolicyCheckData;
			pcdata->url=frameurl;
			pcdata->object = this;

			EvaluatePolicy(pcdata);
		}
	}
	else
	{
		EvaluatePolicy();	
	}


	return S_OK;
}

void CP3PObject::EvaluatePolicy(PolicyCheckData* data, bool waitforthread)
{
	if( !m_bDocumentCompleted )
		return;

	if( !waitforthread && workerthread )
	{
		if( workerthread->wait(1) )  ///check whether alive
			return; ///return if still alive
	}

	PolicyCheckData* pcdata = data;;
	if( !pcdata )
	{
		pcdata = new PolicyCheckData;
		pcdata->url=GetLocationURL();
		pcdata->object = this;
	}

	matchinfo.SetValid(false);

	///if not http hide the p3p 
	GURL &url = pcdata->url;
	if( url.is_empty() || (url.protocol() != "http" && url.protocol() != "https"))
	{
		icondlg.SetImageNoBird();
		return;
	}
	icondlg.Show();
	///////////////////////////////////////////////////////

	if( workerthread )
	{
		while(workerthread->wait(THREAD_WAIT_TIME))
		{
			workerthreadmonitor.enter();
			workerthreadmonitor.broadcast();
			workerthreadmonitor.leave();
		}
		delete workerthread;
		workerthread = 0;
	}
	matchinfo.SetValid(true);
	///get all the embed urls here because it is hanging if it done in other thread
	embed_list.empty();
	GetAllEmbeddedURLs(*m_spWebBrowser2, embed_list);

	TRY
	{
		linkrefurl = GURL();
		linkrefurl = this->FindLinkReferenceInDocument(m_spWebBrowser2);
	}CATCH(ex){}ENDCATCH;

	workerthread = new GThread;
	workerthread->create(PolicyCheckStart, pcdata);
	return;
}


/////////////////////////////////////////////////////////////////////////////
// OnQuit - The browser (and the BHO) is terminating
// 
HRESULT CP3PObject::OnQuit()
{
	// Unregister as an event listener 
	Disconnect();

	

	///kill pending threads
	matchinfo.SetValid(false);

	if( workerthread )
	{
		if(workerthread->wait(THREAD_WAIT_TIME))
		{
			workerthreadmonitor.enter();
			workerthreadmonitor.broadcast();
			workerthreadmonitor.leave();
		}
		//delete workerthread;
		//workerthread = 0;
	}
	// Destroy the Code Window 
	if( icondlg.m_hWnd )
		icondlg.DestroyWindow();
	// Unregister the window hook previously installed

	if( hook)
	{
		if( UnhookWindowsHookEx(hook) )
			hook = 0;
	}
	// Unregister the keyboard hook previously installed
	if( g_hKbHook )
		UnhookWindowsHookEx(g_hKbHook);

	//// remove window procedure from chain
	if( wndproc )
		SetWindowLong(m_hwndBrowser, GWL_WNDPROC, (LONG)(WNDPROC)wndproc);
	

	return S_OK;
}


/////////////////////////////////////////////////////////////////////////////
// DisplayP3PIcon - Display the icon on top of title
// 
BOOL CP3PObject::DisplayP3PIcon()
{
	
	// Unless it's already visible...
	
	if (!IsWindowVisible(icondlg.m_hWnd)) 
	{
		// The Code Window is created as a child of IE
		icondlg.Create(m_hwndBrowser);
		icondlg.Reposition(m_hwndBrowser);
		icondlg.ShowWindow(SW_SHOWNORMAL);
	}


	return true;
}





/////////////////////////////////////////////////////////////////////////////
// RetrieveBrowserWindow - Get the HWND of the browser
// 
HRESULT CP3PObject::RetrieveBrowserWindow()
{
	// To get the HWND of the browser we need to enumerate
	// all the windows belonging to this thread. We could
	// check for the window's class name if it wasn't changed
	// from Windows 95/NT 4.0 to Windows 98. With Windows 98
	// the class name is the same as any other open folder, so
	// there's no certainty of uniqueness.

	// Get the HWND of the browser's window
	EnumThreadWindows( 
		GetCurrentThreadId(),
		WndEnumProc,
		reinterpret_cast<LPARAM>(&m_hwndBrowser) );

	if (!IsWindow(m_hwndBrowser))
		return E_POINTER;

	return S_OK;
}


/////////////////////////////////////////////////////////////////////////////
// WndEnumProc - Enumeration procedure for thread's windows
// 
BOOL CALLBACK CP3PObject::WndEnumProc(HWND hwnd, LPARAM lParam)
{
	TCHAR szClassName[MAX_PATH];
	GetClassName(hwnd, szClassName, MAX_PATH);

	// IEFrame was the window's class name under Win95. The
	// other is the title under Win98.
	if (!lstrcmpi(szClassName, _T("CabinetWClass")) ||
        !lstrcmpi(szClassName, _T("IEFrame")))
	{
		// We need to return the HWND found if any. The lParam
		// is a pointer to the HWND to be used the return buffer
		HWND *phWnd = reinterpret_cast<HWND*>(lParam);
		*phWnd = hwnd;

		// Return FALSE to stop enumeration once we've found the 
		// expected window
		return false;
	}

	// Return TRUE to continue enumerating windows
	return true;
}



/////////////////////////////////////////////////////////////////////////////
// GetCodeWindowRect - Set the rectangle for the Code Window
// 
HRESULT CP3PObject::GetCodeWindowRect(LPRECT lprc)
{
	// Both the rectangle of IE and the Code Window are
	// calculated by the class constructor
	CopyRect(lprc, &m_rcHTML);
	return S_OK;
}


void CP3PObject::RepositionIcon()
{
	icondlg.Reposition(m_hwndBrowser);
}





void CP3PObject::OnBeforeNavigate(IDispatch *pDisp,
    GString &URL,
    DWORD &Flags,
    GString &TargetFrameName,
    GString &PostData,
    GString &Headers,
    BOOL &Cancel)
{
	{	
		if( PostData.length() )
		{
			GURL url = URL;
#ifdef _DEBUG
			FILE *fp;
			if( fp = fopen("c:\\temp\\temp.txt", "a") )
			{
				fprintf(fp, "\n\n%s\n",(const char*)url);
				//fprintf(fp, (const char*) PostData);
				fclose(fp);
			}
#endif

			if( !g_p3p_preferences.GetNoPromptAllForms() && !g_p3p_preferences.IsInNoPromptSiteList(url.host()) )
			{

				Result result(url);
				int retval = PostCheck(GURL(URL), result);

				if( retval != 0 )
				{
					CPostWarningDialog pwd(retval<0?0:&result);
					if( pwd.DoModal() == IDCANCEL )
						Cancel = TRUE;

					if( pwd.GetDontWarnStatus() )
					{
						g_p3p_preferences.AddToNoPromptSiteList(GURL(URL).host(), true);
						SendCommandToAllP3Ps(ID_CHANGED_NOPROMPTSITES);
					}
				}
			}

		}

		if( !Cancel && IsTopFrame(pDisp) )
		{
			locationURL = URL;
			m_bDocumentCompleted = false;
			matchinfo.SetValid(false);
			matchinfo.CleanCookies();
			icondlg.SetImageUnknown();
			if( CM_CleanRequests )
				CM_CleanRequests(GetCurrentThreadId());
			if( PostData.length() )
				isPostRequest=true;
			else
				isPostRequest=false;
		}	
	}	
}


GURL CP3PObject::GetLocationURL()
{
	///////////////////////////////////////////////////////////----------------------
	// Get the WebBrowser's document object
	TRY
	{
		BSTR str=0;
		HRESULT hr = m_spWebBrowser2->get_LocationURL(&str);
		if (FAILED(hr))
			return GURL();

		GString strlocation = cvt_ws_to_utf8(str);
		SysFreeString(str);
		GURL locURL = strlocation;

		return locURL;
	}
	CATCH(ex)
	{
		
	}
	ENDCATCH;
	
	return GURL();
}


int CP3PObject::PostCheck(
    GURL &checkurl,
    Result &result)
{
	int retval = -1;

	if( checkurl == checkurl.host() )
		checkurl = checkurl+"/";

	GURL saveurl = checkurl;
	GURL url = saveurl;
	///if not http return
	if( url.is_empty() || (url.protocol() != "http" && url.protocol() != "https"))
	{
		return retval;
	}

	icondlg.SetImageBusy();

	GString method = "POST";

	if( g_policy_handler.FindPolicy(saveurl, method).is_empty() )
	{
		url = url.host()+"/w3c/p3p.xml";
		g_policy_handler.HandlePolicyReferenceURL(url, saveurl);
	}

	
	TRY
	{
		
		GURL polurl = g_policy_handler.FindPolicy(saveurl, method);
		if( !polurl.is_empty() )
		{
			retval = g_policy_handler.HandlePolicyURL(polurl, url, result);
		}
		else
		{
			GList<GURL> hint_list;
			g_policy_handler.GetApplicableHints(saveurl, hint_list);
			for(GPosition hintpos = hint_list; hintpos && polurl.is_empty(); ++hintpos)
			{
				g_policy_handler.HandlePolicyReferenceURL(hint_list[hintpos], hint_list[hintpos]);
				polurl = g_policy_handler.FindPolicy(saveurl, method);
			}
			if( !polurl.is_empty() )
			{
				retval = g_policy_handler.HandlePolicyURL(polurl, url, result);
			}
		}
		
	}
	CATCH(ex)
	{
		retval = -1;
	}
	ENDCATCH;

	TRY
	{

	/// show the previous result
		bool blocked_cookies;
		switch(matchinfo.GetNetResult(blocked_cookies))
		{
		case P3PMatchInfo::UNKNOWN:
			if( blocked_cookies )
				icondlg.SetImageUnknownBlockedCookies();
			else
				icondlg.SetImageUnknown();
			break;
		case P3PMatchInfo::MATCH:
			if( blocked_cookies )
				icondlg.SetImageMatchBlockedCookies();
			else
				icondlg.SetImageMatch();
			break;
		case P3PMatchInfo::UNMATCH:
			if( blocked_cookies )
				icondlg.SetImageNoMatchBlockedCookies();
			else
				icondlg.SetImageNoMatch();
			break;
		case P3PMatchInfo::MATCH_NOTHIRDPARTY:
			if( blocked_cookies )
				icondlg.SetImageMatch3rdPartyNomatchBlockedCookies();
			else
				icondlg.SetImageMatch3rdPartyNomatch();
			break;
		}
	}
	CATCH(ex)
	{
		icondlg.SetImageUnknown();
	}
	ENDCATCH;

	return retval;
	
}


int CP3PObject::PolicyCheck(
    GURL &checkurl,
    DWORD &Flags,
    GString &TargetFrameName,
    GString &PostData,
    GString &Headers)
{
	int retval = 0;

	if( checkurl.is_empty() )
		return retval;

	GURL saveurl = checkurl;
	GURL url = saveurl;

	try
	{
	
	TRY
	{
		
		matchinfo.CleanInfo(); /////////////////////////////??????????????????????????

		icondlg.SetImageBusy();

		GString method = PostData.length()?"POST":"GET";

		if( g_policy_handler.FindPolicy(saveurl, method).is_empty() )
		{
			url = url.host()+"/w3c/p3p.xml";
			g_policy_handler.HandlePolicyReferenceURL(url, saveurl);
		}

	
		bool fromcache=false;
		GURL polurl = g_policy_handler.FindPolicy(saveurl, method);
		if( !polurl.is_empty() )
		{
			GURL tempurl = polurl;
			SeparateNameAndURL(tempurl);
			if( g_policy_handler.isURLExistsInCache(tempurl) )
				fromcache = true;
			retval = matchinfo.CheckMainURL(saveurl, polurl);
		}
		else
		{
			bool flag = false;
			
			if( !linkrefurl.is_empty() )
			{
				g_policy_handler.HandlePolicyReferenceURL(linkrefurl, saveurl);
				GURL polurl = g_policy_handler.FindPolicy(saveurl, method);
				if( !polurl.is_empty() )
				{
					retval = matchinfo.CheckMainURL(saveurl, polurl);
					flag = true;
				}
			}

			if( !flag && !isPostRequest)
			{
				//GString strurl = saveurl;
				
				//if( strurl.search('?') < 0 )
					retval = matchinfo.CheckMainURLWithP3PHeader(saveurl);
			}
		}

		//////////////////////////////////////////////////test ?????????????????????
		{
			int match = Result::UNKNOWN;

			if( embed_list.size() )
			{
				for( GPosition pos=embed_list; pos; ++pos )
				{
					matchinfo.AddEmbededURL(embed_list[pos].URL, embed_list[pos].type);
					
				}
				matchinfo.CheckEmbeds();
			}

			matchinfo.CheckAllCookiePolicies();

			bool blocked_cookies;
			match = matchinfo.GetNetResult(blocked_cookies);
				
			switch(match)
			{
			default:
			case P3PMatchInfo::UNKNOWN:
				if( blocked_cookies )
					icondlg.SetImageUnknownBlockedCookies();
				else
					icondlg.SetImageUnknown();
				if( (g_p3p_preferences.enable_sounds & P3PPreferences::PLAY_SOUND_CACHED_POLICY) || !fromcache )
					P3PMatchInfo::PlaySoundPolicyUnkown();
				break;
			case P3PMatchInfo::MATCH:
				if( blocked_cookies )
					icondlg.SetImageMatchBlockedCookies();
				else
					icondlg.SetImageMatch();
				if( (g_p3p_preferences.enable_sounds & P3PPreferences::PLAY_SOUND_CACHED_POLICY) || !fromcache )
					P3PMatchInfo::PlaySoundPolicyMatched();
				break;
			case P3PMatchInfo::UNMATCH:
				if( blocked_cookies )
					icondlg.SetImageNoMatchBlockedCookies();
				else
					icondlg.SetImageNoMatch();
				if( (g_p3p_preferences.enable_sounds & P3PPreferences::PLAY_SOUND_CACHED_POLICY) || !fromcache )
					P3PMatchInfo::PlaySoundPolicyUnmatched();
				break;
			case P3PMatchInfo::MATCH_NOTHIRDPARTY:
				if( blocked_cookies )
					icondlg.SetImageMatch3rdPartyNomatchBlockedCookies();
				else
					icondlg.SetImageMatch3rdPartyNomatch();
				if( (g_p3p_preferences.enable_sounds & P3PPreferences::PLAY_SOUND_CACHED_POLICY) || !fromcache )
					P3PMatchInfo::PlaySoundPolicyMatched();
				break;
			}
		}
	}
	CATCH(ex)
	{
		icondlg.SetImageUnknown();
#ifdef _DEBUG
		//MessageBox(this->m_hwndBrowser, ex.get_cause(), "Error", MB_OK);
#endif		retval = -1;
	}
	ENDCATCH;
	}
	catch(char *)
	{
		icondlg.SetImageUnknown();
	}
	catch(...)
	{
		icondlg.SetImageUnknown();
#ifdef _DEBUG
		//MessageBox(this->m_hwndBrowser, "Unknown Bug", "Error", MB_OK);
#endif
		retval = -1;
	}

	return retval;
	
}

GURL CP3PObject::FindLinkReferenceInDocument(CComQIPtr<IWebBrowser2, &IID_IWebBrowser2> m_spWebBrowser2)
{
	if( !m_spWebBrowser2 )
		return GURL();

	CComPtr<IDispatch> pDisp;
	HRESULT hr = m_spWebBrowser2->get_Document(&pDisp);
	if (FAILED(hr))
		return GURL();

	CComQIPtr<IHTMLDocument2, &IID_IHTMLDocument2> htmldoc;
	htmldoc = pDisp;

				
	if( htmldoc )
	{
		
		CComPtr<IHTMLElementCollection> p;
		
		HRESULT hr = htmldoc->get_all(&p);

		if (FAILED(hr))
			return GURL();

		long length=0;
		hr = p->get_length(&length);

		if (FAILED(hr))
			return GURL();

		for(long i=0; i<length; i++)
		{
			CComPtr<IDispatch> pDisp;
			_variant_t vname(i);
			_variant_t vid;
			hr = p->item(vname, vid, &pDisp);
			if (FAILED(hr))
				return GURL();

			CComQIPtr<IHTMLLinkElement, &IID_IHTMLLinkElement> link;
			link = pDisp;
			if( link )
			{
				BSTR brel=0;
				link->get_rel(&brel);
				GString rel = cvt_ws_to_utf8(brel);
				SysFreeString(brel);
				if( rel.upcase() == "P3PV1" )
				{
					BSTR bhref=0;
					link->get_href(&bhref);
					GString href = cvt_ws_to_utf8(bhref);
					SysFreeString(bhref);
					if( href.length() )
					{
						BSTR str=0;
						hr = htmldoc->get_URL(&str);
						if (FAILED(hr))
							return GURL();

						GString strlocation = cvt_ws_to_utf8(str);
						SysFreeString(str);
						GURL locURL = strlocation;
						return GURL(href, locURL);
					}
				}
			}
		}

	}

	return GURL();
}


////////////////////////////////////////////////////////////////////////////
/**************************************************************************/
////////////////////////////////////////////////////////////////////////////

void PolicyCheckStart(void *arg)
{
	PolicyCheckData *pcdata = (PolicyCheckData*)arg;
	pcdata->object->AddWorkerThreadMonitor(GetCurrentThreadId());
	try
	{
	pcdata->object->PolicyCheck(pcdata->url, pcdata->flags, pcdata->TargetFrameName, pcdata->PostData, pcdata->Headers);
	}
	catch(...)
	{
	}
	pcdata->object->DeleteWorkerThreadMonitor(GetCurrentThreadId());
	delete pcdata;

}


void CP3PObject::GetAllEmbeddedURLs(IWebBrowser2 &browser, GList<P3PMatchInfo::URLTypeInfo> &list)
{

	////Get the urls first
	CComPtr<IDispatch> pDisp;
	HRESULT hr = browser.get_Document(&pDisp);

	if( FAILED(hr) )
		return;

	if (pDisp)
	{
		CComQIPtr<IHTMLDocument2, &IID_IHTMLDocument2> spHTML;
		spHTML = pDisp;
		if( spHTML )
		{	
			GetAllEmbeddedURLs(spHTML, list);
		}
	}

	try
	{

	if( pDisp )
	{
		IOleContainer* pContainer=0;		

 		hr = pDisp->QueryInterface(IID_IOleContainer, (void**)&pContainer);

		if( FAILED(hr) || !pContainer)
			return;
		
		///Get Frame enumarator
		CComPtr<IEnumUnknown> pEnumerator;
		try
		{

		
		hr = pContainer->EnumObjects(OLECONTF_EMBEDDINGS, &pEnumerator);
		}
		catch(...)
		{
#ifdef WANTCATCHMESSAGES	
			MessageBox(this->m_hwndBrowser, "Unknown Bug in pContainer->EnumObjects", "Error", MB_OK);
#endif

		}
		if( pContainer )
			pContainer->Release();

		if( FAILED(hr) || !pEnumerator)
			return;
		
		try
		{

			///enumerate frames
			for(UINT i=0; 1 ; i++)
			{
				CComPtr<IUnknown> pUnk;
				ULONG uFetched;
				if( pEnumerator->Next(1, &pUnk, &uFetched) != S_OK)
					break;
				
				if( pUnk )
				{
					CComPtr<IWebBrowser2> pWebBrowser;
					hr = pUnk->QueryInterface(IID_IWebBrowser2, (void**)&pWebBrowser);
					
					if( SUCCEEDED(hr) && pWebBrowser)
					{
						
						BSTR str=0;
						HRESULT hr = pWebBrowser->get_LocationURL(&str);
						if (SUCCEEDED(hr))
						{
							GString strlocation = cvt_ws_to_utf8(str);
							SysFreeString(str);
							GString src = strlocation;
							if( src.length() )
								list.append(P3PMatchInfo::URLTypeInfo("Frame/IFrame", GURL(src)));
						}
						
						
						GetAllEmbeddedURLs(*pWebBrowser, list);
					}
				}
			}
		}
		catch(...)
		{
#ifdef _DEBUG			
			MessageBox(this->m_hwndBrowser, "Unknown Bug in EmbedURLs in frame 3", "Error", MB_OK);
#endif

		}
	}

	}
	catch(...)
	{
#ifdef _DEBUG
		MessageBox(this->m_hwndBrowser, "Unknown Bug in EmbedURLs in frame", "Error", MB_OK);
#endif
	}

}



void CP3PObject::GetAllEmbeddedURLs(CComQIPtr<IHTMLDocument2, &IID_IHTMLDocument2> htmldoc, GList<P3PMatchInfo::URLTypeInfo> &list)
{
	if( htmldoc )
	{
		BSTR durl=0;
		HRESULT hr = htmldoc->get_URL(&durl);
		if (FAILED(hr))
			return;

		GURL docurl(cvt_ws_to_utf8(durl));
		SysFreeString(durl);
		{
			CComPtr<IHTMLElementCollection> p;
		
			//// get all the images
			hr = htmldoc->get_all(&p);

			if (SUCCEEDED(hr))
			{
				long length=0;
				hr = p->get_length(&length);

				if (SUCCEEDED(hr))
				{
					for(long i=0; i<length; i++)
					{
						CComPtr<IDispatch> pDisp;
						_variant_t vname(i);
						_variant_t vid;
						hr = p->item(vname, vid, &pDisp);
						if (SUCCEEDED(hr))
						{
							GString name;
							try
							{

							CComQIPtr<IHTMLElement, &IID_IHTMLElement> item;
							item = pDisp;
							if( item )
							{
								BSTR bname=0;
								item->get_tagName(&bname);
								name = cvt_ws_to_utf8(bname).upcase();
								SysFreeString(bname);

								GString src;
								GString type;
								if( name == "IMG" )
								{
									CComQIPtr<IHTMLImgElement, &IID_IHTMLImgElement> imgitem;
									imgitem = item;
									if( imgitem )
									{
										BSTR bhref=0;
										imgitem->get_href(&bhref);
										src = cvt_ws_to_utf8(bhref);
										SysFreeString(bhref);
										type = "Image";
									}
								}
								
								else if( name == "EMBED" )
								{
									CComQIPtr<IHTMLEmbedElement, &IID_IHTMLEmbedElement> embeditem;
									embeditem = item;
									if( embeditem )
									{
										BSTR bsrc=0;
										embeditem->get_src(&bsrc);
										src = cvt_ws_to_utf8(bsrc);
										SysFreeString(bsrc);
										type = "Embed";
									}
								}
								/*
								else if( name == "IFRAME" )
								{
									VARIANT val;
									VariantClear(&val);
									_bstr_t attrname("src");
									if( item->getAttribute(attrname, false, &val) == S_OK)
									{
										if( val.vt == VT_BSTR )
										{
											src = cvt_ws_to_utf8(val.bstrVal);
											type = "IFrame";
										}
									}
									VariantClear(&val);
								}
								else if( name == "FRAME" )
								{
									VARIANT val;
									VariantClear(&val);
									_bstr_t attrname("src");
									if( item->getAttribute(attrname, false, &val) == S_OK)
									{
										if( val.vt == VT_BSTR )
										{
											src = cvt_ws_to_utf8(val.bstrVal);
											type = "Frame";
										}
									}
									VariantClear(&val);
								}
								*/
								else if( name == "APPLET" )
								{

									VARIANT val;
									_bstr_t attrname("code");
									VariantClear(&val);
									if( item->getAttribute(attrname, false, &val) == S_OK)
									{
										if( val.vt == VT_BSTR )
										{
											src = cvt_ws_to_utf8(val.bstrVal);
											type = "Applet";

											_bstr_t bcodebase("codebase");
											VariantClear(&val);
											if( item->getAttribute(attrname, false, &val) == S_OK)
											{
												if( val.vt == VT_BSTR )
												{
													GString base = cvt_ws_to_utf8(val.bstrVal);
													src = base + src;
												}
											}
										}
									}
									VariantClear(&val);

								}
								else if( name == "OBJECT" )
								{
									CComQIPtr<IHTMLObjectElement, &IID_IHTMLObjectElement> objectitem;
									objectitem = item;
									if(objectitem)
									{
										BSTR bcode=0;
										if( SUCCEEDED(objectitem->get_code(&bcode) ) )
										{
											src = cvt_ws_to_utf8(bcode);
											SysFreeString(bcode);
											type = "Object";

											BSTR bcodebase=0;
											if( SUCCEEDED(objectitem->get_codeBase(&bcodebase) ) )
											{
												GString base = cvt_ws_to_utf8(bcodebase);
												SysFreeString(bcodebase);
												if(src.length())
													src = base+"/"+src;
												else
													src = base;
											}
										}
									}
								}
								else 
								{
									CComQIPtr<IHTMLLinkElement, &IID_IHTMLLinkElement> linkitem;
									linkitem = item;
									if( linkitem )
									{
										BSTR brel=0;
										linkitem->get_rel(&brel);
										GString rel = cvt_ws_to_utf8(brel);
										SysFreeString(brel);
										if( rel.upcase() == "STYLESHEET" )
										{
											BSTR bhref=0;
											linkitem->get_href(&bhref);
											src = cvt_ws_to_utf8(bhref);
											SysFreeString(bhref);
											type = "StyleSheet";
										}
									}
								}



								TRY
								{
								if( src.length() && type.length() )
									list.append(P3PMatchInfo::URLTypeInfo(type, GURL(src, docurl)));
								}
								CATCH(ex)
								{}
								ENDCATCH;
								
							}
							}
							catch(...)
							{
#ifdef WANTCATCHMESSAGES	
								MessageBox(this->m_hwndBrowser, name+":Unknown Bug in EmbedURLs", "Error", MB_OK);
#endif
							}
						}
					}
				}
			}
		}		
	}
	return;
}

bool CP3PObject::IsTopFrame(IDispatch *pDisp)
{
	CComPtr<IUnknown> lpUnknownWB;
	CComPtr<IUnknown> lpUnknownFrame;

	HRESULT hr = this->m_spWebBrowser2->QueryInterface(IID_IUnknown, (LPVOID*)&lpUnknownWB);
	if( SUCCEEDED(hr) )
	{
		hr = pDisp->QueryInterface(IID_IUnknown, (LPVOID*)&lpUnknownFrame);
		if( SUCCEEDED(hr) && lpUnknownFrame == lpUnknownWB )
		{
			return true;
		}
	}
	return false;
}

bool CP3PObject::AddCookie(const char *name, const char *value, const char *domain, const char *trackerdomain, const char *path)
{
	return matchinfo.AddCookie(name, value, domain, trackerdomain, path, locationURL);
}


bool CP3PObject::NeedCookieBlocking(const char *name, const char *value, const char *domain, const char *trackerdomain, const char *path)
{
	return matchinfo.NeedCookieBlocking(name, value, domain, trackerdomain, path, locationURL);
}

void CP3PObject::InformRedirection(const char *OriginURL, const char *RedirectedURL)
{
	if( GString(locationURL) == OriginURL )
	{
		TRY
		{
			locationURL = RedirectedURL;
		}
		CATCH(ex)
		{
			locationURL = OriginURL;
		}
		ENDCATCH;

	}
}



/*  End of module: P3PObject.cpp  */
