#include "stdafx.h"
#include "P3PPreferences.h"
#include "XmlFile.h"
#include "XmlNode.h"
#include <time.h>
#include "Globals.h"
#include "resource.h"
#include "GOS.h"
#include "SaxXmlParse.h"

#define DEFAULT_PRIVACY_LEVEL P3PPreferences::MEDIUM  
#define DEFAULT_CHECK_UPGRADE false  
#define DEFAULT_ENABLE_SOUNDS ENABLE_SOUND_UNMATCHED|ENABLE_SOUND_MATCHED
#define DEFAULT_DONTWARN_FORMS false 
#define DEFAULT_P3P_ENABLED true 
#define DEFAULT_HOTKEY_MOD 3
#define DEFAULT_HOTKEY_KEY VK_F12
#define DEFAULT_USERPREFS	0

P3PPreferences g_p3p_preferences;

GCriticalSection P3PPreferences::class_lock;

GMap<GString, GMap<GString, bool> > P3PPreferences::cookie_history;


int P3PPreferences::Load()
{
	InitializationLoad(g_dllpath+INITIALIZATION_FILE);
	CookieLoad(g_dllpath+COOKIE_FILE);
	GetCurPreferenceFile(true);
	NoPromptSitesLoad(g_dllpath+NO_PROMPT_SITES_FILE);
	LoadCookieHistory();

	return 1;
}


int P3PPreferences::InitializationLoad(const char* lpszPathName)
{
	GCriticalSectionLock lock(&class_lock);

	if( lpszPathName )
	{
		FILE *fp;
		if( fp = fopen(lpszPathName, "rb") )
		{
			int value1=DEFAULT_PRIVACY_LEVEL, value2=DEFAULT_CHECK_UPGRADE;
			int value3=DEFAULT_ENABLE_SOUNDS, value4=DEFAULT_DONTWARN_FORMS;
			int value5=DEFAULT_P3P_ENABLED;
			int value6=DEFAULT_HOTKEY_MOD, value7=DEFAULT_HOTKEY_KEY;
			int value8=DEFAULT_USERPREFS;

			fseek(fp, 0, SEEK_SET);
			fscanf( fp, "privacy_level=%d\n", &value1 );
			if( (value1 >= P3PPreferences::LOW) && (value1 <= P3PPreferences::IMPORTED_PRIVACY) )
				privacy_level = (PrivacyLevels) value1;
			else
				privacy_level = (PrivacyLevels) DEFAULT_PRIVACY_LEVEL;

			if( (privacy_level == IMPORTED_PRIVACY) && (!GOS::is_file(g_dllpath+IMPORTED_PREFERENCE_FILE)) )
				privacy_level = (PrivacyLevels) DEFAULT_PRIVACY_LEVEL;

			/*
			fscanf( fp, "check_upgrade=%d\n", &value2 );
			if( (value2==0) || (value2==1) )
				check_upgrade = value2 ? 1:0;
			else
				check_upgrade = DEFAULT_CHECK_UPGRADE;
			*/

			fscanf( fp, "enable_sounds=%d\n", &value3 );
			if( (value3 >= 0) && (value3 <= 7) )
				enable_sounds = value3;
			else
				enable_sounds = DEFAULT_ENABLE_SOUNDS;

			fscanf( fp, "dontwarn_forms=%d\n", &value4 );
			if( (value4 == 0) || (value4 == 1) )
				dontwarn_forms = value4?true:false;
			else
				dontwarn_forms = DEFAULT_DONTWARN_FORMS;

			fscanf( fp, "enabled=%d\n", &value5 );
			if( (value5 == 0) || (value5 == 1) )
				g_enabled = value5?true:false;
			else
				g_enabled = DEFAULT_P3P_ENABLED;

			fscanf( fp, "hotkey_mod=%d\n", &value6 );
			if( (value6 >= 1) && (value6 <= 7) )
				hotkey_mod = value6;
			else
				hotkey_mod = DEFAULT_HOTKEY_MOD;

			fscanf( fp, "hotkey_key=%d\n", &value7 );
			if( (value7 >= 1)  )
				hotkey_key = value7;
			else
				hotkey_key = DEFAULT_HOTKEY_MOD;

			fscanf(fp, "user_prefs=%d\n", &value8 );
			if( value8 >= 1 )
				user_prefs = value8;
			else
				user_prefs = DEFAULT_USERPREFS;

			g_onTitleBar = true;
			fscanf(fp, "on_titlebar=%d\n", &g_onTitleBar);
			g_Position.x = 0; g_Position.y = 0;
			fscanf(fp, "position=%d, %d\n", &g_Position.x, &g_Position.y);

			RECT rect;
			SystemParametersInfo(SPI_GETWORKAREA, 0, &rect, 0);
			//int cx = GetSystemMetrics(SM_CXFULLSCREEN);
		//	int cy = GetSystemMetrics(SM_CYFULLSCREEN);
			if( g_Position.x < rect.left )
				g_Position.x = rect.left;
			if( g_Position.x > rect.right )
				g_Position.x = rect.right-50;
			if( g_Position.y < rect.top )
				g_Position.y = rect.top;
			if( g_Position.y > rect.bottom )
				g_Position.y = rect.bottom-10;

			fclose(fp);
			return 1;
		}
	}

	// In case the filename is empty or file can't be opened or the expected value
	// is not found in the file, then initialize to the default level(=Medium).
	privacy_level = (PrivacyLevels) DEFAULT_PRIVACY_LEVEL;
	check_upgrade = (bool) DEFAULT_CHECK_UPGRADE;
	enable_sounds =  DEFAULT_ENABLE_SOUNDS;
	user_prefs = DEFAULT_USERPREFS;

	return 0;
}

	
int P3PPreferences::CookieLoad(const char* lpszPathName)
{
	GCriticalSectionLock lock(&class_lock);

	TRY
	{
		this->domain_prefs.empty();

		GP<XmlNode> root = SaxXmlParse(lpszPathName, false);//XMLParse(lpszPathName);
		
		if( root && root->name == "COOKIERULESET" )
		{
			GPosition pos = root->first_child();
			while( pos )
			{
				GP<XmlNode> childnode = root->next_child(pos);

				//// Load the site-specific (i.e., override) Cookie Settings 
				if( childnode->name == "COOKIERULE" )
				{
					GString domain;
					GString overridecookiesetting;

					for( GPosition attr_pos=childnode->attrvalues; attr_pos; ++attr_pos )
					{
						GString key = childnode->attrvalues.key(attr_pos);
						GString value = childnode->attrvalues[attr_pos];
						if( key == "DOMAIN" )
						{
							domain = value;
						}
						else if( key == "ACCEPT" )
						{
							overridecookiesetting= value;
						}
					}

					if( overridecookiesetting.length() )
					{
						site_allow_cookies prefs;

						if( overridecookiesetting == "ALL" )
						{
							prefs = ALWAYS;
							if( domain.length() )
								this->domain_prefs[domain] = prefs;
						}
						else if( overridecookiesetting == "NONE")
						{
							prefs = NEVER;
							if( domain.length() )
								this->domain_prefs[domain] = prefs;
						}
				
					
					}
				}

				//// Load the general (i.e., default to all sites) Cookie Settings 
				else if( childnode->name == "DEFAULTCOOKIERULE" )
				{
					GString acceptcookie, headers, cross_site;

					for( GPosition attr_pos=childnode->attrvalues; attr_pos; ++attr_pos )
					{
						GString key = childnode->attrvalues.key(attr_pos);
						GString value = childnode->attrvalues[attr_pos].upcase();
						if( key == "ACCEPT" )
						{
							acceptcookie= value;
						}
						else if( key == "HEADERS" )
						{
							headers= value;
						}
						else if( key == "CROSS-SITE" )
						{
							cross_site= value;
						}
					}

					if( headers.length() || cross_site.length() )
					{
						this->allow_cookies = CUSTOM_COOKIE;

						if( acceptcookie == "ALL" )
							this->allow_cookies_custom = ALL;
						else if( acceptcookie == "NONE")
							this->allow_cookies_custom = NONE;
						else 
							this->allow_cookies_custom = MATCHING_PREFERENCES;
					
						if( headers == "ALL" )
							this->allow_cookies_custom_lastpagevisited = ALL;
						else if( headers == "NONE")
							this->allow_cookies_custom_lastpagevisited = NONE;
						else 
							this->allow_cookies_custom_lastpagevisited = MATCHING_PREFERENCES;

						if( cross_site == "ALL" )
							this->allow_cookies_custom_crosssite = ALL;
						else if( cross_site == "NONE")
							this->allow_cookies_custom_crosssite = NONE;
						else 
							this->allow_cookies_custom_crosssite = MATCHING_PREFERENCES;
					}
					else
					{
						if( acceptcookie == "ALL" )
							this->allow_cookies = ALL;
						else if( acceptcookie == "NONE")
							this->allow_cookies = NONE;
						else 
							this->allow_cookies = MATCHING_PREFERENCES;
					}
				}

			}
		}

		return 1;
	}
	CATCH(ex)
	{
	}
	ENDCATCH;

	return 0;
}


bool P3PPreferences::ReadPreferenceFile(DArray<char> &data)
{
	GCriticalSectionLock lock(&class_lock);

	FILE *fp;
	GString filename = GetCurPreferenceFile(false);
	if( filename.length() > 0 )
	{
		if( fp = fopen(filename, "rb") )
		{
			fseek(fp, 0, SEEK_END);
			int len = ftell(fp);
			fseek(fp, 0, SEEK_SET);
			
			data.resize(0, len-1);
			fread(data, len, 1, fp);

			fclose(fp);
			return true;
		}
	}

	return false;
}

GString P3PPreferences::GetCurPreferenceFile(bool Reload)
{
	GString preference_filename;
	switch (privacy_level)
	{
		case P3PPreferences::LOW :
			preference_filename = LOW_PREFERENCE_FILE; 
			break;

		case P3PPreferences::MEDIUM :
			preference_filename = MEDIUM_PREFERENCE_FILE; 
			break;

		case P3PPreferences::HIGH :
			preference_filename = HIGH_PREFERENCE_FILE; 
			break;

		case P3PPreferences::CUSTOM_PRIVACY :
			preference_filename = PREFERENCE_FILE; 
			break;

		case P3PPreferences::IMPORTED_PRIVACY :
			preference_filename = IMPORTED_PREFERENCE_FILE; 
			break;

	}

	preference_filename = g_dllpath+preference_filename;
	if( Reload )
		PreferenceLoad(preference_filename);

	return( preference_filename );
}


int P3PPreferences::PreferenceLoad(const char* lpszPathName)
{

	GCriticalSectionLock lock(&class_lock);

	warn_medical_analysis = warn_medical_share = false;
	warn_financial_analysis = warn_financial_share = false;
	warn_nonpersonal_analysis = warn_nonpersonal_share = false;
	warn_personal_analysis = warn_personal_share = false;
	warn_contact_phone = warn_contact_mail = false;
	warn_no_remove_mailinglist = warn_no_access_data = false;

	if( privacy_level == IMPORTED_PRIVACY )
		return 1;

	TRY
	{
		GP<XmlNode> root = SaxXmlParse(lpszPathName, false);//XMLParse(lpszPathName);
		
		if( root && root->name == "APPEL:RULESET" )
		{
			GPosition pos = root->first_child();
			while( pos )
			{
				GP<XmlNode> childnode = root->next_child(pos);
				if( childnode->name == "APPEL:RULE" )
				{
					ParseRuleElement(childnode);
				}
			}
		}

		return 1;
	}
	CATCH(ex)
	{
	}
	ENDCATCH;
	return 0;
}


int P3PPreferences::ParseRuleElement(GP<XmlNode> node)
{
	if( node && node->name == "APPEL:RULE" )
	{
		if( GPosition pos = node->attrvalues.contains("BEHAVIOR") )
		{
			GString behavior = node->attrvalues[pos].upcase();
			if( (behavior == "LIMITED") || (behavior == "BLOCK") )
			{
				GPosition pos = node->first_child();
				while( pos )
				{
					GP<XmlNode> childnode = node->next_child(pos);
					if( childnode->name == "APPEL:REQUEST-GROUP" )
					{
						//ParseRequestGroupElement(childnode);
					}
					else if( childnode->name == "P3P:POLICY" )
					{
						ParsePolicyElement(childnode);
					}
				}
			}
		}
		return 1;
	}

	return 0;
}


int P3PPreferences::ParsePolicyElement(GP<XmlNode> node)
{
	if( node && node->name == "P3P:POLICY" )
	{
		GPosition pos = node->first_child();
		while( pos )
		{
			GP<XmlNode> childnode = node->next_child(pos);
			if( childnode->name == "P3P:STATEMENT" )
			{
				ParseStatementElement(childnode);
			}
			else if( childnode->name == "P3P:ACCESS" )
			{
				ParseAccessElement(childnode);
			}
		}

		return 1;
	}

	return 0;
}


int P3PPreferences::ParseAccessElement(GP<XmlNode> node)
{
	if( node && node->name == "P3P:ACCESS" )
	{
		GPosition pos = node->first_child();
		while( pos )
		{
			GP<XmlNode> childnode = node->next_child(pos);
			if( childnode->name == "P3P:NONE" )
			{
				warn_no_access_data = true;
			}
		}

		return 1;
	}

	return 0;
}

/*
int P3PPreferences::ParseStatementElement(GP<XmlNode> node)
{

	if( node && node->name == "P3P:STATEMENT" )
	{
		bool multiple_condition = false;
		if( GPosition attr_pos = node->attrvalues.contains("CONNECTIVE") )
		{
			GString connective = node->attrvalues[attr_pos].upcase();
			if( connective == "AND" )
				multiple_condition= true;
		}

		int found_recipient_group = 0;
		bool save_warn_nonpersonal_share = warn_nonpersonal_share;
		warn_nonpersonal_share = false;

		GPosition pos = node->first_child();
		while( pos )
		{
			GP<XmlNode> childnode = node->next_child(pos);
			if( childnode->name == "P3P:PURPOSE" )
			{
				ParsePurposeElement(childnode);
			}
			else if( childnode->name == "P3P:RECIPIENT" )
			{
				ParseRecipientElement(childnode);

				found_recipient_group++;
			}
			else if( childnode->name == "P3P:DATA-GROUP" )
			{
				ParseDataGroupElement(childnode);
				if( found_recipient_group && warn_nonpersonal_share && categories_personal && multiple_condition)
				{
					warn_personal_share = true;
				}
				found_recipient_group = 0;
			}
		}

		if( !found_recipient_group )
			warn_nonpersonal_share = save_warn_nonpersonal_share;


		return 1;
	}

	return 0;
}*/


int P3PPreferences::ParseStatementElement(GP<XmlNode> node)
{

	if( node && node->name == "P3P:STATEMENT" )
	{
		if( GPosition attr_pos = node->attrvalues.contains("APPEL:CONNECTIVE") )
		{
			GString connective = node->attrvalues[attr_pos].upcase();
			if( connective != "AND" )
				return 0;
		}

		recipient_thirdparty = false;
		purpose_pseudo_analysis = purpose_pseudo_decision = false;
		purpose_individual_analysis = purpose_individual_decision = false;
		purpose_contact_always = purpose_contact_optout = false;
		purpose_telemarketing_always = purpose_telemarketing_optout = false;
		categories_personal = categories_medical = categories_financial = false;

		GPosition pos = node->first_child();
		while( pos )
		{
			GP<XmlNode> childnode = node->next_child(pos);
			if( childnode->name == "P3P:PURPOSE" )
			{
				ParsePurposeElement(childnode);
			}
			else if( childnode->name == "P3P:RECIPIENT" )
			{
				ParseRecipientElement(childnode);
			}
			else if( childnode->name == "P3P:DATA-GROUP" )
			{
				ParseDataGroupElement(childnode);
			}
		}


		bool pseudo = purpose_pseudo_analysis || purpose_pseudo_decision;
		bool individual = purpose_individual_analysis || purpose_individual_decision;
		bool contact = purpose_contact_always || purpose_contact_optout;
		bool telemarketing = purpose_telemarketing_always || purpose_telemarketing_optout;
		bool analysis = pseudo || individual || contact || telemarketing;

		if( recipient_thirdparty )
		{
			if( categories_personal )
				warn_personal_share = true;
			if( categories_financial )
				warn_financial_share = true;
			if( categories_medical )
				warn_medical_share = true;
			if(! (categories_personal || categories_financial || categories_medical) )
				warn_nonpersonal_share = true;
		}

		if( categories_financial && analysis)
			warn_financial_analysis = true;
		if( categories_medical && analysis)
			warn_medical_analysis = true;
		if(! (categories_financial || categories_medical) )
		{
			if( pseudo )
				warn_nonpersonal_analysis = true;
			if( individual )
				warn_personal_analysis = true;
			if( purpose_telemarketing_always && purpose_contact_always )
				warn_no_remove_mailinglist = true;
			else 
			{
				if( purpose_telemarketing_always || purpose_telemarketing_optout)
					warn_contact_phone = true;
				if( purpose_contact_always || purpose_contact_optout )
					warn_contact_mail = true;
			}
		}

		return 1;
	}

	return 0;
}


int P3PPreferences::ParsePurposeElement(GP<XmlNode> node)
{
	if( node && node->name == "P3P:PURPOSE" )
	{
		if( GPosition attr_pos = node->attrvalues.contains("APPEL:CONNECTIVE") )
		{
			GString required = node->attrvalues[attr_pos].upcase();
			if( required != "OR")
				return 0;
		}

		GPosition pos = node->first_child();
		while( pos )
		{
			GP<XmlNode> childnode = node->next_child(pos);

			if( childnode->name == "P3P:PSEUDO-ANALYSIS" )
			{
				if( GPosition pos = childnode->attrvalues.contains("REQUIRED") )
				{
					GString required = childnode->attrvalues[pos].upcase();
					if( (required == "ALWAYS") || (required == "OPT-OUT") )
						purpose_pseudo_analysis = true;
				}
				else
					purpose_pseudo_analysis = true;
			}
			else if( childnode->name == "P3P:PSEUDO-DECISION" )
			{
				if( GPosition pos = childnode->attrvalues.contains("REQUIRED") )
				{
					GString required = childnode->attrvalues[pos].upcase();
					if( (required == "ALWAYS") || (required == "OPT-OUT") )
						purpose_pseudo_decision = true;
				}
				else
					purpose_pseudo_decision = true;
			}
			else if( childnode->name == "P3P:INDIVIDUAL-ANALYSIS" )
			{
				if( GPosition pos = childnode->attrvalues.contains("REQUIRED") )
				{
					GString required = childnode->attrvalues[pos].upcase();
					if( (required == "ALWAYS") || (required == "OPT-OUT") )
						purpose_individual_analysis = true;
				}
				else
					purpose_individual_analysis = true;
			}
			else if( childnode->name == "P3P:INDIVIDUAL-DECISION" )
			{
				if( GPosition pos = childnode->attrvalues.contains("REQUIRED") )
				{
					GString required = childnode->attrvalues[pos].upcase();
					if( (required == "ALWAYS") || (required == "OPT-OUT") )
						purpose_individual_decision = true;
				}
				else
					purpose_individual_decision = true;
			}
			else if( childnode->name == "P3P:CONTACT" )
			{
				if( GPosition pos = childnode->attrvalues.contains("REQUIRED") )
				{
					GString required = childnode->attrvalues[pos].upcase();
					if( required == "ALWAYS") 
					{
						purpose_contact_always = true;
					}
					else if( required == "OPT-OUT")
					{
						purpose_contact_optout = true;
					}
				}

			}
			else if( childnode->name == "P3P:TELEMARKETING" )
			{
				// if the "Required" attribute is absent, or set to value "=always"
				// then, it implies there is no option to remove from telemarketing list.
				if( GPosition pos = childnode->attrvalues.contains("REQUIRED") )
				{
					GString required = childnode->attrvalues[pos].upcase();
					if( required == "ALWAYS") 
					{
						purpose_telemarketing_always = true;
					}
					else if( required == "OPT-OUT")
					{
						purpose_telemarketing_optout = true;
					}

				}

			}
		}

		return 1;
	}

	return 0;
}


int P3PPreferences::ParseRecipientElement(GP<XmlNode> node)
{
	if( node && node->name == "P3P:RECIPIENT" )
	{
		if( GPosition attr_pos = node->attrvalues.contains("APPEL:CONNECTIVE") )
		{
			GString required = node->attrvalues[attr_pos].upcase();
			if( required != "OR")
				return 0;
		}

		GPosition pos = node->first_child();
		while( pos )
		{
			GP<XmlNode> childnode = node->next_child(pos);

			if( (childnode->name == "P3P:SAME") 
				|| (childnode->name == "P3P:OTHER-RECIPIENT") 
				|| (childnode->name == "P3P:UNRELATED")  
				|| (childnode->name == "P3P:PUBLIC") )
			{
				if( GPosition pos = childnode->attrvalues.contains("REQUIRED") )
				{
					GString required = childnode->attrvalues[pos].upcase();
					if( (required == "ALWAYS") || (required == "OPT-OUT") )
						recipient_thirdparty= true;
				}
				else
					recipient_thirdparty= true;
			}
		}

		return 1;
	}

	return 0;
}


int P3PPreferences::ParseDataGroupElement(GP<XmlNode> node)
{
	if( node && node->name == "P3P:DATA-GROUP" )
	{
		GPosition pos = node->first_child();
		while( pos )
		{
			GP<XmlNode> childnode = node->next_child(pos);
			if( childnode->name == "P3P:DATA" )
			{
				ParseDataElement(childnode);
			}
		}

		return 1;
	}

	return 0;
}


int P3PPreferences::ParseDataElement(GP<XmlNode> node)
{
	if( node && node->name == "P3P:DATA" )
	{
		GPosition pos = node->first_child();
		while( pos )
		{
			GP<XmlNode> childnode = node->next_child(pos);
			if( childnode->name == "P3P:CATEGORIES" )
			{
				ParseCategoriesElement(childnode);
			}
		}

		return 1;
	}

	return 0;
}


int P3PPreferences::ParseCategoriesElement(GP<XmlNode> node)
{
	if( node && node->name == "P3P:CATEGORIES" )
	{
		if( GPosition attr_pos = node->attrvalues.contains("APPEL:CONNECTIVE") )
		{
			GString required = node->attrvalues[attr_pos].upcase();
			if( required != "OR")
				return 0;
		}

		GPosition pos = node->first_child();
		while( pos )
		{
			GP<XmlNode> childnode = node->next_child(pos);
			if( childnode->name == "P3P:HEALTH" )
			{
				categories_medical = true;
			}
			else if( childnode->name == "P3P:FINANCIAL" || childnode->name == "P3P:PURCHASE" )
			{
				categories_financial = true;
			}
			else if( childnode->name == "P3P:PHYSICAL" )
			{
				categories_personal = true;
			}
			else if( childnode->name == "P3P:ONLINE" )
			{
				categories_personal = true;
			}
			else if( childnode->name == "P3P:GOVERNMENT" )
			{
				categories_personal = true;
			}
		}

		return 1;
	}

	return 0;
}


int P3PPreferences::Save()
{

	InitializationSave(g_dllpath+INITIALIZATION_FILE);
	CookieSave(g_dllpath+COOKIE_FILE);
	PreferenceSave(g_dllpath+PREFERENCE_FILE);
	SaveCookieHistory();

	return 1;
}


int P3PPreferences::InitializationSave(const char* lpszPathName)
{
	GCriticalSectionLock lock(&class_lock);

	if( lpszPathName )
	{
		FILE *fp = fopen(lpszPathName, "w");
		if( !fp )
			THROW("Could not open file for writing, File could be locked by other program");

		fprintf(fp, "privacy_level=%d\n", (int)this->privacy_level);
		//fprintf(fp, "check_upgrade=%d\n", (int)this->check_upgrade);
		fprintf(fp, "enable_sounds=%d\n", this->enable_sounds);
		fprintf(fp, "dontwarn_forms=%d\n", this->dontwarn_forms);
		fprintf(fp, "enabled=%d\n", g_enabled);
		fprintf(fp, "hotkey_mod=%d\n", hotkey_mod);
		fprintf(fp, "hotkey_key=%d\n", hotkey_key);
		fprintf(fp, "user_prefs=%d\n", user_prefs);
		fprintf(fp, "on_titlebar=%d\n", g_onTitleBar);
		fprintf(fp, "position=%d, %d\n", g_Position.x, g_Position.y);
		fclose(fp);
	}
	else
		THROW("Empty Initialization file name");

	return 1;
}


int P3PPreferences::CookieSave(const char* lpszPathName)
{
	GCriticalSectionLock lock(&class_lock);

	TRY
	{
		XmlOutFile xmlf(lpszPathName);


		//// Save the Cookie Rule Set  
		XmlOutNode ruleset(xmlf);
		ruleset.setNodeName("COOKIERULESET", atts_empty, vals_empty);
			

		{
			//// Save the site-specific (i.e., override) Cookie Settings 

			GString overridecookiesetting="";

			for( GPosition pos=this->domain_prefs; pos; ++pos)
			{/// iterate through all URLs to be saved. 

				site_allow_cookies prefs = this->domain_prefs[pos];
				GString domain = this->domain_prefs.key(pos);

				switch (prefs)
				{
					case P3PPreferences::ALWAYS :
						overridecookiesetting = "ALL"; 
						break;

					case P3PPreferences::NEVER :
						overridecookiesetting = "NONE"; 
						break;
				
					default :
						overridecookiesetting = ""; 
						break;
				}
				
				//save the settings only if different from default.
				if (overridecookiesetting.length())  
				{
					GList<GString> atts;
					GList<GString> vals;
					atts.append("domain");
					vals.append(domain);
					atts.append("accept");
					vals.append(overridecookiesetting);

					XmlOutNode site_rule(xmlf, true);
					site_rule.setNodeName("COOKIERULE", atts, vals);
				}
			
			}

		}			
		
		{
			//// Save the general (default for all sites) Cookie Settings 

			GString acceptcookie, headers, cross_site;

			switch (this->allow_cookies)
			{
				case P3PPreferences::MATCHING_PREFERENCES :
					acceptcookie = "MATCHING"; 
					break;

				case P3PPreferences::ALL :
					acceptcookie = "ALL"; 
					break;

				case P3PPreferences::NONE :
					acceptcookie = "NONE"; 
					break;

				case P3PPreferences::CUSTOM_COOKIE :
					{ 	switch (this->allow_cookies_custom)
						{
						case P3PPreferences::MATCHING_PREFERENCES :
							acceptcookie = "MATCHING"; break;

						case P3PPreferences::ALL :
							acceptcookie = "ALL"; break;

						case P3PPreferences::NONE :
							acceptcookie = "NONE"; break;

						}

						switch (this->allow_cookies_custom_lastpagevisited)
						{
						case P3PPreferences::MATCHING_PREFERENCES :
							headers = "MATCHING"; break;

						case P3PPreferences::ALL :
							headers = "ALL"; break;

						case P3PPreferences::NONE :
							headers = "NONE"; break;

						}

						switch (this->allow_cookies_custom_crosssite)
						{
						case P3PPreferences::MATCHING_PREFERENCES :
							cross_site = "MATCHING"; break;

						case P3PPreferences::ALL :
							cross_site = "ALL"; break;

						case P3PPreferences::NONE :
							cross_site = "NONE"; break;

						}
					}
			
			}  //end  of switch()
			
			GList<GString> atts;
			GList<GString> vals;
			atts.append("accept");
			vals.append(acceptcookie);

			if (headers.length())
			{
				atts.append("headers");
				vals.append(headers);
			}
			if (cross_site.length())
			{
				atts.append("cross-site");
				vals.append(cross_site);
			}

			XmlOutNode general_rule(xmlf, true);
			general_rule.setNodeName("DEFAULTCOOKIERULE", atts, vals);

		}

	}
	CATCH(ex)
	{
		
	}
	ENDCATCH;

	return 1;
}



int P3PPreferences::PreferenceSave(const char* lpszPathName)
{
	GCriticalSectionLock lock(&class_lock);

	TRY
	{
		XmlOutFile xmlf(lpszPathName);

		//// write all data related to nodes
		{
			XmlOutNode ruleset(xmlf);
			GList<GString> atts;
			GList<GString> vals;
			atts.append("xmlns:appel");
			vals.append("http://www.w3.org/2001/02/APPELv1");
			atts.append("xmlns:p3p");
			vals.append("http://www.w3.org/2000/12/P3Pv1");
			atts.append("crtdby");
			vals.append("AT&amp;T Privacy Bird");
			atts.append("crtdon");
			time_t ltime;
			time(&ltime);
			vals.append(ctime(&ltime)); 
			ruleset.setNodeName("appel:RULESET", atts, vals);

			if( warn_medical_analysis )
			{
				MedicalAnalysisAlways(xmlf);
				MedicalAnalysisOptout(xmlf);
			}
			if( warn_medical_share )
			{
				MedicalShareAlways(xmlf);
				MedicalShareOptout(xmlf);
			}
			if( warn_financial_analysis )
			{
				FinancialAnalysisAlways(xmlf);
				FinancialAnalysisOptout(xmlf);
			}
			if( warn_financial_share )
			{
				FinancialShareAlways(xmlf);
				FinancialShareOptout(xmlf);
			}

			if( warn_contact_phone )
			{
				ContactPhoneAlways(xmlf);
				ContactPhoneOptout(xmlf);
			}
			if( warn_contact_mail )
			{
				ContactMailAlways(xmlf);
				ContactMailOptout(xmlf);
			}
			if( warn_no_remove_mailinglist )
			{
				NoRemoveMailingList(xmlf);
			}
			if( warn_personal_analysis )
			{
				PersonalAnalysisAlways(xmlf);
				PersonalAnalysisOptout(xmlf);
			}
			if( warn_personal_share )
			{
				PersonalShareAlways(xmlf);
				PersonalShareOptout(xmlf);
			}
			if( warn_no_access_data )
			{
				NoAccessData(xmlf);
			}

			if( warn_nonpersonal_analysis )
			{
				NonPersonalAnalysisAlways(xmlf);
				NonPersonalAnalysisOptout(xmlf);
			}
			if( warn_nonpersonal_share )
			{
				NonPersonalShareAlways(xmlf);
				NonPersonalShareOptout(xmlf);
			}
			
			UnknownPurposeAlways(xmlf);
			UnknownPurposeOptout(xmlf);
			//xmlf.print("\n");
			Otherwise(xmlf);
		}
	}
	CATCH(ex)
	{
		
	}
	ENDCATCH;

	return 1;
}



int P3PPreferences::MedicalAnalysisAlways(XmlOutFile &xmlf)
{
	{
		XmlOutNode rule(xmlf);
		GList<GString> atts;
		GList<GString> vals;
		atts.append("behavior");
		vals.append("limited");
		atts.append("description");
		vals.append("Site may use health or medical information for analysis "
					"or to make decisions that may affect what content "
					"or ads you see, etc.");
		rule.setNodeName("appel:RULE", atts, vals);

		XmlOutNode policy(xmlf);
		policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

		XmlOutNode statement(xmlf);
		statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

		PurposeAnalysis(xmlf, vals_always, true, true, true, true, false, false);
		DataGroup(xmlf, true, false, false);
	}

	{
		XmlOutNode rule(xmlf);
		GList<GString> atts;
		GList<GString> vals;
		atts.append("behavior");
		vals.append("limited");
		atts.append("description");
		vals.append("Site may use health or medical information for marketing");
		rule.setNodeName("appel:RULE", atts, vals);

		XmlOutNode policy(xmlf);
		policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

		XmlOutNode statement(xmlf);
		statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

		PurposeAnalysis(xmlf, vals_always, false, false, false, false, true, true);
		DataGroup(xmlf, true, false, false);
	}

	return 1;
}


int P3PPreferences::MedicalAnalysisOptout(XmlOutFile &xmlf)
{
	{
		XmlOutNode rule(xmlf);
		GList<GString> atts;
		GList<GString> vals;
		atts.append("behavior");
		vals.append("limited");
		atts.append("description");
		vals.append("Unless you opt-out, site may use health or medical information "
					"for analysis or to make decisions that may affect "
					"what content or ads you see, etc.");
		rule.setNodeName("appel:RULE", atts, vals);

		XmlOutNode policy(xmlf);
		policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

		XmlOutNode statement(xmlf);
		statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

		PurposeAnalysis(xmlf, vals_optout, true, true, true, true, false, false);
		DataGroup(xmlf, true, false, false);
	}

	{
		XmlOutNode rule(xmlf);
		GList<GString> atts;
		GList<GString> vals;
		atts.append("behavior");
		vals.append("limited");
		atts.append("description");
		vals.append("Unless you opt-out, site may use health or medical information "
					"for marketing");
		rule.setNodeName("appel:RULE", atts, vals);

		XmlOutNode policy(xmlf);
		policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

		XmlOutNode statement(xmlf);
		statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

		PurposeAnalysis(xmlf, vals_optout, false, false, false, false, true, true);
		DataGroup(xmlf, true, false, false);
	}
	
	return 1;
}

int P3PPreferences::MedicalShareAlways(XmlOutFile &xmlf)
{
	XmlOutNode rule(xmlf);
	GList<GString> atts;
	GList<GString> vals;
	atts.append("behavior");
	vals.append("limited");
	atts.append("description");
	vals.append("Site may share health or medical information "
				"with other companies (other than those helping "
				"the site provide services to you)");
	rule.setNodeName("appel:RULE", atts, vals);

	XmlOutNode policy(xmlf);
	policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

	XmlOutNode statement(xmlf);
	statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

	RecipientShare(xmlf, vals_always);
	DataGroup(xmlf, true, false, false);

	return 1;
}


int P3PPreferences::MedicalShareOptout(XmlOutFile &xmlf)
{
	XmlOutNode rule(xmlf);
	GList<GString> atts;
	GList<GString> vals;
	atts.append("behavior");
	vals.append("limited");
	atts.append("description");
	vals.append("Unless you opt-out, site may share health or medical information "
				"with other companies (other than those helping "
				"the site provide services to you)");
	rule.setNodeName("appel:RULE", atts, vals);

	XmlOutNode policy(xmlf);
	policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

	XmlOutNode statement(xmlf);
	statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

	RecipientShare(xmlf, vals_optout);
	DataGroup(xmlf, true, false, false);

	return 1;
}


int P3PPreferences::FinancialAnalysisAlways(XmlOutFile &xmlf)
{
	{
		XmlOutNode rule(xmlf);
		GList<GString> atts;
		GList<GString> vals;
		atts.append("behavior");
		vals.append("limited");
		atts.append("description");
		vals.append("Site may use financial information or information about your "
					"purchases for analysis or to make decisions that may affect "
					"what content or ads you see, etc.");
		rule.setNodeName("appel:RULE", atts, vals);

		XmlOutNode policy(xmlf);
		policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

		XmlOutNode statement(xmlf);
		statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

		PurposeAnalysis(xmlf, vals_always, true, true, true, true, false, false);
		DataGroup(xmlf, false, true, false);
	}

	{
		XmlOutNode rule(xmlf);
		GList<GString> atts;
		GList<GString> vals;
		atts.append("behavior");
		vals.append("limited");
		atts.append("description");
		vals.append("Site may use financial information or information about your "
					"purchases for marketing");
		rule.setNodeName("appel:RULE", atts, vals);

		XmlOutNode policy(xmlf);
		policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

		XmlOutNode statement(xmlf);
		statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

		PurposeAnalysis(xmlf, vals_always, false, false, false, false, true, true);
		DataGroup(xmlf, false, true, false);
	}
	
	return 1;
}


int P3PPreferences::FinancialAnalysisOptout(XmlOutFile &xmlf)
{
	{
		XmlOutNode rule(xmlf);
		GList<GString> atts;
		GList<GString> vals;
		atts.append("behavior");
		vals.append("limited");
		atts.append("description");
		vals.append("Unless you opt-out, site may use financial information or information "
					"about your purchases for analysis or to make decisions "
					"that may affect what content or ads you see, etc.");
		rule.setNodeName("appel:RULE", atts, vals);

		XmlOutNode policy(xmlf);
		policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

		XmlOutNode statement(xmlf);
		statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

		PurposeAnalysis(xmlf, vals_optout, true, true, true, true, false, false);
		DataGroup(xmlf, false, true, false);
	}

	{
		XmlOutNode rule(xmlf);
		GList<GString> atts;
		GList<GString> vals;
		atts.append("behavior");
		vals.append("limited");
		atts.append("description");
		vals.append("Unless you opt-out, site may use financial information or information "
					"about your purchases for marketing");
		rule.setNodeName("appel:RULE", atts, vals);

		XmlOutNode policy(xmlf);
		policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

		XmlOutNode statement(xmlf);
		statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

		PurposeAnalysis(xmlf, vals_optout, false, false, false, false, true, true);
		DataGroup(xmlf, false, true, false);
	}

	return 1;
}


int P3PPreferences::FinancialShareAlways(XmlOutFile &xmlf)
{
	XmlOutNode rule(xmlf);
	GList<GString> atts;
	GList<GString> vals;
	atts.append("behavior");
	vals.append("limited");
	atts.append("description");
	vals.append("Site may share financial information or information about your "
				"purchases with other companies (other than those helping "
				"the site provide services to you)");
	rule.setNodeName("appel:RULE", atts, vals);

	XmlOutNode policy(xmlf);
	policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

	XmlOutNode statement(xmlf);
	statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

	RecipientShare(xmlf, vals_always);
	DataGroup(xmlf, false, true, false);

	return 1;
}


int P3PPreferences::FinancialShareOptout(XmlOutFile &xmlf)
{
	XmlOutNode rule(xmlf);
	GList<GString> atts;
	GList<GString> vals;
	atts.append("behavior");
	vals.append("limited");
	atts.append("description");
	vals.append("Unless you opt-out, site may share financial information or "
				"information about your purchases with other companies (other than "
				"those helping the site provide services to you)");
	rule.setNodeName("appel:RULE", atts, vals);

	XmlOutNode policy(xmlf);
	policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

	XmlOutNode statement(xmlf);
	statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

	RecipientShare(xmlf, vals_optout);
	DataGroup(xmlf, false, true, false);

	return 1;
}


int P3PPreferences::ContactMailAlways(XmlOutFile &xmlf)
{
	XmlOutNode rule(xmlf);
	GList<GString> atts;
	GList<GString> vals;
	atts.append("behavior");
	vals.append("limited");
	atts.append("description");
	vals.append("Site may contact you through means other "
	           "than telephone (email, postal mail, etc.) to try to "
	           "interest you in other services or products");
	rule.setNodeName("appel:RULE", atts, vals);

	XmlOutNode policy(xmlf);
	policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

	XmlOutNode statement(xmlf);
	statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

	PurposeAnalysis(xmlf, vals_always, false, false, false, false, true, false);

	return 1;
}


int P3PPreferences::ContactMailOptout(XmlOutFile &xmlf)
{
	XmlOutNode rule(xmlf);
	GList<GString> atts;
	GList<GString> vals;
	atts.append("behavior");
	vals.append("limited");
	atts.append("description");
	vals.append("Unless you opt-out, site may contact you through "
	           "means other than telephone (email, postal mail, etc.) "
	           "to interest you in other services or products");
	rule.setNodeName("appel:RULE", atts, vals);

	XmlOutNode policy(xmlf);
	policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

	XmlOutNode statement(xmlf);
	statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

	PurposeAnalysis(xmlf, vals_optout, false, false, false, false, true, false);

	return 1;
}


int P3PPreferences::ContactPhoneAlways(XmlOutFile &xmlf)
{
	XmlOutNode rule(xmlf);
	GList<GString> atts;
	GList<GString> vals;
	atts.append("behavior");
	vals.append("limited");
	atts.append("description");
	vals.append("Site may contact you via telephone to interest "
	           "you in other services or products");
	rule.setNodeName("appel:RULE", atts, vals);

	XmlOutNode policy(xmlf);
	policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

	XmlOutNode statement(xmlf);
	statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

	PurposeAnalysis(xmlf, vals_always, false, false, false, false, false, true);

	return 1;
}


int P3PPreferences::ContactPhoneOptout(XmlOutFile &xmlf)
{
	XmlOutNode rule(xmlf);
	GList<GString> atts;
	GList<GString> vals;
	atts.append("behavior");
	vals.append("limited");
	atts.append("description");
	vals.append("Unless you opt-out, site may contact you via telephone "
	           "to interest you in other services or products");
	rule.setNodeName("appel:RULE", atts, vals);

	XmlOutNode policy(xmlf);
	policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

	XmlOutNode statement(xmlf);
	statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

	PurposeAnalysis(xmlf, vals_optout, false, false, false, false, false, true);

	return 1;
}


int P3PPreferences::NoRemoveMailingList(XmlOutFile &xmlf)
{
	XmlOutNode rule(xmlf);
	GList<GString> atts;
	GList<GString> vals;
	atts.append("behavior");
	vals.append("limited");
	atts.append("description");
	vals.append("Site may contact you to interest you in other "
		        "services or products and does not allow you to "
			    "remove yourself from marketing/mailing list");
	rule.setNodeName("appel:RULE", atts, vals);

	XmlOutNode policy(xmlf);
	policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

	XmlOutNode statement(xmlf);
	statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

	PurposeAnalysis(xmlf, vals_always, false, false, false, false, true, true);

	return 1;
}


int P3PPreferences::PersonalAnalysisAlways(XmlOutFile &xmlf)
{
	{
		XmlOutNode rule(xmlf);
		GList<GString> atts;
		GList<GString> vals;
		atts.append("behavior");
		vals.append("limited");
		atts.append("description");
		vals.append("Site may keep information that personally identifies you to "
				  "determine your habits, interests, or other characteristics "
				  "for research and analysis purposes");
		rule.setNodeName("appel:RULE", atts, vals);

		XmlOutNode policy(xmlf);
		policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

		XmlOutNode statement(xmlf);
		statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

		PurposeAnalysis(xmlf, vals_always, false, false, true, false, false, false);
	}

	{
		XmlOutNode rule(xmlf);
		GList<GString> atts;
		GList<GString> vals;
		atts.append("behavior");
		vals.append("limited");
		atts.append("description");
		vals.append("Site may keep information that personally identifies you to "
				  "determine your habits, interests, or other characteristics for "
				  "making decisions about what content or ads you see at the site, etc.");
		rule.setNodeName("appel:RULE", atts, vals);

		XmlOutNode policy(xmlf);
		policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

		XmlOutNode statement(xmlf);
		statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

		PurposeAnalysis(xmlf, vals_always, false, false, false, true, false, false);
	}

	return 1;
}


int P3PPreferences::PersonalAnalysisOptout(XmlOutFile &xmlf)
{
	{
		XmlOutNode rule(xmlf);
		GList<GString> atts;
		GList<GString> vals;
		atts.append("behavior");
		vals.append("limited");
		atts.append("description");
		vals.append("Unless you opt-out, site may keep information that personally identifies "
				  "you to determine your habits, interests, or other characteristics for "
				  "research and analysis purposes");
		rule.setNodeName("appel:RULE", atts, vals);

		XmlOutNode policy(xmlf);
		policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

		XmlOutNode statement(xmlf);
		statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

		PurposeAnalysis(xmlf, vals_optout, false, false, true, false, false, false);
	}

	{
		XmlOutNode rule(xmlf);
		GList<GString> atts;
		GList<GString> vals;
		atts.append("behavior");
		vals.append("limited");
		atts.append("description");
		vals.append("Unless you opt-out, site may keep information that personally identifies "
				  "you to determine your habits, interests, or other characteristics for "
				  "making decisions about what content or ads you see at the site, etc.");
		rule.setNodeName("appel:RULE", atts, vals);

		XmlOutNode policy(xmlf);
		policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

		XmlOutNode statement(xmlf);
		statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

		PurposeAnalysis(xmlf, vals_optout, false, false, false, true, false, false);
	}

	return 1;
}


int P3PPreferences::PersonalShareAlways(XmlOutFile &xmlf)
{
	XmlOutNode rule(xmlf);
	GList<GString> atts;
	GList<GString> vals;
	atts.append("behavior");
	vals.append("limited");
	atts.append("description");
	vals.append("Site may share information that personally identifies you "
				"with other companies (other than those helping "
				"the site provide services to you)");
	rule.setNodeName("appel:RULE", atts, vals);

	XmlOutNode policy(xmlf);
	policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

	atts.empty();
	vals.empty();

	XmlOutNode statement(xmlf);
	atts.append("appel:connective");
	vals.append("and");
	statement.setNodeName("p3p:STATEMENT", atts, vals);

	RecipientShare(xmlf, vals_always);
	DataGroup(xmlf, false, false, true);

	return 1;
}


int P3PPreferences::PersonalShareOptout(XmlOutFile &xmlf)
{
	XmlOutNode rule(xmlf);
	GList<GString> atts;
	GList<GString> vals;
	atts.append("behavior");
	vals.append("limited");
	atts.append("description");
	vals.append("Unless you opt-out, site may share information that personally "
				"identifies you with other companies (other than those helping "
				"the site provide services to you)");
	rule.setNodeName("appel:RULE", atts, vals);

	XmlOutNode policy(xmlf);
	policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

	atts.empty();
	vals.empty();

	XmlOutNode statement(xmlf);
	atts.append("appel:connective");
	vals.append("and");
	statement.setNodeName("p3p:STATEMENT", atts, vals);

	RecipientShare(xmlf, vals_optout);
	DataGroup(xmlf, false, false, true);

	return 1;
}


int P3PPreferences::NoAccessData(XmlOutFile &xmlf)
{
	XmlOutNode rule(xmlf);
	GList<GString> atts;
	GList<GString> vals;
	atts.append("behavior");
	vals.append("limited");
	atts.append("description");
	vals.append("Site does not allow you to find out "
			    "what data they have about you");
	rule.setNodeName("appel:RULE", atts, vals);

	XmlOutNode policy(xmlf);
	policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

	XmlOutNode access(xmlf);
	access.setNodeName("p3p:ACCESS", atts_empty, vals_empty);

	XmlOutNode none(xmlf, true);
	none.setNodeName("p3p:none", atts_empty, vals_empty);

	return 1;
}


int P3PPreferences::NonPersonalAnalysisAlways(XmlOutFile &xmlf)
{
	{
		XmlOutNode rule(xmlf);
		GList<GString> atts;
		GList<GString> vals;
		atts.append("behavior");
		vals.append("limited");
		atts.append("description");
		vals.append("Site may keep information that does not personally identify you "
				  "to determine your habits, interests, or other characteristics "
				  "for research and analysis purposes");
		rule.setNodeName("appel:RULE", atts, vals);

		XmlOutNode policy(xmlf);
		policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

		XmlOutNode statement(xmlf);
		statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

		PurposeAnalysis(xmlf, vals_always, true, false, false, false, false, false);
	}

	{
		XmlOutNode rule(xmlf);
		GList<GString> atts;
		GList<GString> vals;
		atts.append("behavior");
		vals.append("limited");
		atts.append("description");
		vals.append("Site may keep information that does not personally identify you "
				  "to determine your habits, interests, or other characteristics for "
				  "making decisions about what content or ads you see at the site, etc.");
		rule.setNodeName("appel:RULE", atts, vals);

		XmlOutNode policy(xmlf);
		policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

		XmlOutNode statement(xmlf);
		statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

		PurposeAnalysis(xmlf, vals_always, false, true, false, false, false, false);
	}

	return 1;
}


int P3PPreferences::NonPersonalAnalysisOptout(XmlOutFile &xmlf)
{
	{
		XmlOutNode rule(xmlf);
		GList<GString> atts;
		GList<GString> vals;
		atts.append("behavior");
		vals.append("limited");
		atts.append("description");
		vals.append("Unless you opt-out, site may keep information "
					"that does not personally identify you to determine your "
					"habits, interests, or other characteristics "
					"for research and analysis purposes");
		rule.setNodeName("appel:RULE", atts, vals);

		XmlOutNode policy(xmlf);
		policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

		XmlOutNode statement(xmlf);
		statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

		PurposeAnalysis(xmlf, vals_optout, true, false, false, false, false, false);
	}
	{
		XmlOutNode rule(xmlf);
		GList<GString> atts;
		GList<GString> vals;
		atts.append("behavior");
		vals.append("limited");
		atts.append("description");
		vals.append("Unless you opt-out, site may keep information "
					"that does not personally identify you to determine your "
					"habits, interests, or other characteristics for "
					"making decisions about what content or ads you see at the site, etc.");

		rule.setNodeName("appel:RULE", atts, vals);

		XmlOutNode policy(xmlf);
		policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

		XmlOutNode statement(xmlf);
		statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

		PurposeAnalysis(xmlf, vals_optout, false, true, false, false, false, false);
	}


	return 1;
}

int P3PPreferences::NonPersonalShareAlways(XmlOutFile &xmlf)
{
	XmlOutNode rule(xmlf);
	GList<GString> atts;
	GList<GString> vals;
	atts.append("behavior");
	vals.append("limited");
	atts.append("description");
	vals.append("Site may share information that does not personally identify you "
				"with other companies (other than those helping "
				"the site provide services to you)");
	rule.setNodeName("appel:RULE", atts, vals);

	XmlOutNode policy(xmlf);
	policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

	XmlOutNode statement(xmlf);
	statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

	RecipientShare(xmlf, vals_always);

	return 1;
}


int P3PPreferences::NonPersonalShareOptout(XmlOutFile &xmlf)
{
	XmlOutNode rule(xmlf);
	GList<GString> atts;
	GList<GString> vals;
	atts.append("behavior");
	vals.append("limited");
	atts.append("description");
	vals.append("Unless you opt-out, site may share information that does not "
				"personally identify you with other companies (other than "
				"those helping the site provide services to you)");
	rule.setNodeName("appel:RULE", atts, vals);

	XmlOutNode policy(xmlf);
	policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

	XmlOutNode statement(xmlf);
	statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

	RecipientShare(xmlf, vals_optout);

	return 1;
}


int P3PPreferences::UnknownPurposeAlways(XmlOutFile &xmlf)
{
	XmlOutNode rule(xmlf);
	GList<GString> atts;
	GList<GString> vals;
	atts.append("behavior");
	vals.append("limited");
	atts.append("description");
	vals.append("This site collects data for an unknown purpose");
	rule.setNodeName("appel:RULE", atts, vals);

	XmlOutNode policy(xmlf);
	policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

	XmlOutNode statement(xmlf);
	statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

	XmlOutNode purpose(xmlf);
	purpose.setNodeName("p3p:PURPOSE", atts_empty, vals_empty);

	XmlOutNode other_purpose_always(xmlf, true);
	other_purpose_always.setNodeName("p3p:other-purpose", atts_reqd, vals_always);

	return 1;
}


int P3PPreferences::UnknownPurposeOptout(XmlOutFile &xmlf)
{
	XmlOutNode rule(xmlf);
	GList<GString> atts;
	GList<GString> vals;
	atts.append("behavior");
	vals.append("limited");
	atts.append("description");
	vals.append("Unless you opt-out, this site collects data for an unknown purpose");
	rule.setNodeName("appel:RULE", atts, vals);

	XmlOutNode policy(xmlf);
	policy.setNodeName("p3p:POLICY", atts_empty, vals_empty);

	XmlOutNode statement(xmlf);
	statement.setNodeName("p3p:STATEMENT", atts_empty, vals_empty);

	XmlOutNode purpose(xmlf);
	purpose.setNodeName("p3p:PURPOSE", atts_empty, vals_empty);

	XmlOutNode other_purpose_optout(xmlf, true);
	other_purpose_optout.setNodeName("p3p:other-purpose", atts_reqd, vals_optout);

	return 1;
}


int P3PPreferences::Otherwise(XmlOutFile &xmlf)
{
	XmlOutNode rule(xmlf);
	GList<GString> atts;
	GList<GString> vals;
	atts.append("behavior");
	vals.append("request");
	rule.setNodeName("appel:RULE", atts, vals);

	XmlOutNode none(xmlf, true);
	none.setNodeName("appel:OTHERWISE", atts_empty, vals_empty);

	return 1;
}


int P3PPreferences::RecipientShare(XmlOutFile &xmlf, GList<GString> vals)
{ 
	//Recipient section

	XmlOutNode recipient(xmlf);
	recipient.setNodeName("p3p:RECIPIENT", atts_appel_connective, vals_or);

	{
		XmlOutNode same_always(xmlf, true);
		same_always.setNodeName("p3p:same", atts_reqd, vals);
	}
	{
		XmlOutNode other_recipient_always(xmlf, true);
		other_recipient_always.setNodeName("p3p:other-recipient", atts_reqd, vals);
	}
	{
		XmlOutNode unrelated_always(xmlf, true);
		unrelated_always.setNodeName("p3p:unrelated", atts_reqd, vals);
	}
	{
		XmlOutNode public_always(xmlf, true);
		public_always.setNodeName("p3p:public", atts_reqd, vals);
	}

	return 1;
}  


int P3PPreferences::DataGroup(XmlOutFile &xmlf, bool medical, bool financial, bool personal_identification)
{ 
	// Data-Group section

	XmlOutNode data_group(xmlf);
	data_group.setNodeName("p3p:DATA-GROUP", atts_empty, vals_empty);

	XmlOutNode data(xmlf);
	data.setNodeName("p3p:DATA", atts_empty, vals_empty);

	int num_conditions =0;
	if( medical ) 
		num_conditions++;
	if( financial ) 
		num_conditions+=2;
	if( personal_identification ) 
		num_conditions+=3;
		
	XmlOutNode categories(xmlf);
	if( num_conditions > 1 )
		categories.setNodeName("p3p:CATEGORIES", atts_appel_connective, vals_or);
	else
		categories.setNodeName("p3p:CATEGORIES", atts_empty, vals_empty);
	
	if( medical)
	{
		{
			XmlOutNode medical(xmlf, true);
			medical.setNodeName("p3p:health", atts_empty, vals_empty);
		}
	}

	if( financial)
	{
		{
			XmlOutNode purchase(xmlf, true);
			purchase.setNodeName("p3p:purchase", atts_empty, vals_empty);
		}
		{
			XmlOutNode financial(xmlf, true);
			financial.setNodeName("p3p:financial", atts_empty, vals_empty);
		}
	}

	if( personal_identification)
	{
		{
			XmlOutNode physical(xmlf, true);
			physical.setNodeName("p3p:physical", atts_empty, vals_empty);
		}
		{
			XmlOutNode online(xmlf, true);
			online.setNodeName("p3p:online", atts_empty, vals_empty);
		}
		{
			XmlOutNode government(xmlf, true);
			government.setNodeName("p3p:government", atts_empty, vals_empty);
		}
	}


	return 1;
} 


int P3PPreferences::PurposeAnalysis(XmlOutFile &xmlf, GList<GString> vals, bool pseudo_analysis, bool pseudo_decision, bool individual_analysis, bool individual_decision, bool contact, bool telemarketing)
{ 
	//Purpose Section

	XmlOutNode purpose(xmlf);
	
	int num_conditions =0;
	if( pseudo_analysis ) 
		num_conditions++;
	if( pseudo_decision ) 
		num_conditions++;
	if( individual_analysis ) 
		num_conditions++;
	if( individual_decision ) 
		num_conditions++;
	if( contact ) 
		num_conditions++;
	if( telemarketing ) 
		num_conditions++;
		
	if( num_conditions > 1 )
		purpose.setNodeName("p3p:PURPOSE", atts_appel_connective, vals_or);
	else
		purpose.setNodeName("p3p:PURPOSE", atts_empty, vals_empty);

	if( pseudo_analysis )
	{
		XmlOutNode pseudo_analysis(xmlf, true);
		pseudo_analysis.setNodeName("p3p:pseudo-analysis", atts_reqd, vals);
	}

	if( pseudo_decision )
	{
		XmlOutNode pseudo_decision(xmlf, true);
		pseudo_decision.setNodeName("p3p:pseudo-decision", atts_reqd, vals);
	}

	if( individual_analysis )
	{
		XmlOutNode individual_analysis(xmlf, true);
		individual_analysis.setNodeName("p3p:individual-analysis", atts_reqd, vals);
	}

	if( individual_decision )
	{
		XmlOutNode individual_decision(xmlf, true);
		individual_decision.setNodeName("p3p:individual-decision", atts_reqd, vals);
	}

	if( contact )
	{
			XmlOutNode contact(xmlf, true);
			contact.setNodeName("p3p:contact", atts_reqd, vals);
	}

	if( telemarketing )
	{
			XmlOutNode telemarketing(xmlf, true);
			telemarketing.setNodeName("p3p:telemarketing", atts_reqd, vals);
	}

	return 1;
}


int P3PPreferences::NoPromptSitesLoad(const char* lpszPathName)
{
	GCriticalSectionLock lock(&class_lock);

	no_prompt_sites_list.empty(); 

	if( lpszPathName )
	{
		FILE *fp;
		if( fp = fopen(lpszPathName, "r") )
		{
			char url[2048]; 
			while(1)
			{
				
				if( fscanf( fp, "%s\n", url ) == EOF )
					break;
				TRY
				{
					no_prompt_sites_list[url] = true;
				}
				CATCH(ex)
				{
				}
				ENDCATCH;
			}
			fclose(fp);
			return 1;
		}
	}

	return 0;
}

	
int P3PPreferences::NoPromptSitesSave(const char* lpszPathName)
{
	GCriticalSectionLock lock(&class_lock);

	if( lpszPathName )
	{
		FILE *fp = fopen(lpszPathName, "w");
		if( fp )
		{
			for( GPosition pos=no_prompt_sites_list; pos; ++pos)
			{ 
				//	if( g_dontwarnsitelist[pos] )
					fprintf(fp, "%s\n", (const char *)no_prompt_sites_list.key(pos));
			}
			fclose(fp);
		}
	}

	return 1;
}


GList<GURL> P3PPreferences::GetNoPromptSiteList()
{
	GCriticalSectionLock lock(&class_lock);
	GList<GURL> url_list;

	for( GPosition pos=no_prompt_sites_list; pos; ++pos)
	{ 
		//	if( g_dontwarnsitelist[pos] )
			url_list.append(no_prompt_sites_list.key(pos));
	}
	return url_list;
}


int P3PPreferences::SetNoPromptSiteList(GList<GURL> url_list )
{
	GCriticalSectionLock lock(&class_lock);
	no_prompt_sites_list.empty(); 
	for( GPosition pos=url_list; pos; ++pos)
	{ 
		GURL url = url_list[pos]; 
		no_prompt_sites_list[url] = true;
	}
	NoPromptSitesSave(g_dllpath+NO_PROMPT_SITES_FILE);
	return 1;
}

int P3PPreferences::GetNoPromptSiteListSize()
{
	GCriticalSectionLock lock(&class_lock);
	return no_prompt_sites_list.size() ;
}

bool P3PPreferences::IsInNoPromptSiteList(GURL url)
{
	GCriticalSectionLock lock(&class_lock);
	return ( no_prompt_sites_list.contains(url)?true:false );
}

void P3PPreferences::AddToNoPromptSiteList(GURL url, bool flag)
{
	GCriticalSectionLock lock(&class_lock);
	no_prompt_sites_list[url] = flag;
	NoPromptSitesSave(g_dllpath+NO_PROMPT_SITES_FILE);
}

void P3PPreferences::AddCookieHistory(const char *sitedomain, const char*trackerdomain)
{
///put in the history
	GCriticalSectionLock lock(&class_lock);
	if( sitedomain && strlen(sitedomain))
	{
		cookie_history[sitedomain];
		if( trackerdomain && strlen(trackerdomain) && strcmp(sitedomain, trackerdomain))
			cookie_history[sitedomain][trackerdomain]=false;
	}
}


void P3PPreferences::LoadCookieHistory()
{
	GCriticalSectionLock lock(&class_lock);
	TRY
	{
		GP<XmlNode> root = SaxXmlParse(g_dllpath+COOKIE_HISTORY_FILE, false);//XMLParse(g_dllpath+COOKIE_HISTORY_FILE);
		if( root && root->name == "COOKIEHISTORY" )
		{
			GPosition pos;
			GP<XmlNode> node = root->get_first_child(pos);
			while(node)
			{
				if( node->name == "MAINSITE" )
				{
					if( GPosition namepos = node->attrvalues.contains("NAME") )
					{
						GString name = node->attrvalues[namepos];
						if( name.length() )
						{
							cookie_history[name];
							GPosition child_pos;
							GP<XmlNode> child_node = node->get_first_child(child_pos);
							while(child_node)
							{
								if( child_node->name == "OTHERSITE" )
								{
									if( GPosition namepos = child_node->attrvalues.contains("NAME") )
									{
										GString othersitename = child_node->attrvalues[namepos];
										if( othersitename.length() )
											cookie_history[name][othersitename] = false;
									}
								}
								child_node = node->get_next_child(child_pos);
							}
						}
					}
				}
				node = root->get_next_child(pos);
			}
		}
	}
	CATCH(ex)
	{
	}
	ENDCATCH;
}
			
void P3PPreferences::SaveCookieHistory()
{
	////load first to merge changes
	LoadCookieHistory();
	GCriticalSectionLock lock(&class_lock);

	TRY
	{
		XmlOutFile xmlf(g_dllpath+COOKIE_HISTORY_FILE);
		//// Save the Cookie Rule Set  
		XmlOutNode cookiehistory(xmlf);
		cookiehistory.setNodeName("COOKIEHISTORY");
		for( GPosition pos = cookie_history; pos; ++pos)
		{
			GMap<GString, GString> values;
			values["NAME"] = cookie_history.key(pos);
			XmlOutNode mainsite(xmlf);
			mainsite.setNodeName("MAINSITE",values);

			GMap<GString, bool> &othersites = cookie_history[pos]; 
			for(GPosition otherpos = othersites; otherpos; ++otherpos)
			{
				GMap<GString, GString> values;
				values["NAME"] = othersites.key(otherpos);
				XmlOutNode othersite(xmlf, true);
				othersite.setNodeName("OTHERSITE",values);
			}
		}
	}
	CATCH(ex)
	{
	}
	ENDCATCH;
}


int P3PPreferences::SetNoPromptAllForms(bool flag )
{
	GCriticalSectionLock lock(&class_lock);
	dontwarn_forms = flag;
	InitializationSave(g_dllpath+INITIALIZATION_FILE);
	SendCommandToAllP3Ps(ID_CHANGED_INIT);
	return 1;
}

bool P3PPreferences::GetNoPromptAllForms()
{
	GCriticalSectionLock lock(&class_lock);
	return dontwarn_forms;
}
