#include <stdio.h>
#include <string.h>
#include "P3PEvaluator.h"
#include "XmlNode.h"
#include "Match.h"
#include "PolicySummary.h"
#include <stdio.h>
#include <iostream.h>
#include "PatMatch.h"
#include "PolicyHandler.h"
#include "SaxXmlParse.h"


//    simplify data defs
GString P3PEvaluator::PrepareNewDataDefName(const char *defname,
					    const char *refstructname,
					    const char *matchstructname) {
  int i;
  
  for(i=0;
      refstructname[i] &&
	matchstructname[i] &&
	refstructname[i] == matchstructname[i]; i++)
    continue;
  return GString(defname)+&matchstructname[i];
}

int P3PEvaluator::SimplifyDataDefs(GP<XmlNode> dataschema) {
  // returns 1 if simplified else 0 used for recurtion inside 
  if( dataschema && dataschema->name == "DATASCHEMA" ) {
    // prepare list of data struct for easy searching
    GMap<GString, GP<XmlNode> > datastructs;
    GList<GP<XmlNode> > datadefs;
    GPosition pos;
    GP<XmlNode> node = dataschema->get_first_child(pos);
    while(node) {
      if( node->name == "DATA-STRUCT" ) {
	if( GPosition namepos = node->attrvalues.contains("NAME") ) {
	  datastructs[node->attrvalues[namepos]] = node;
	}
      }
      else if( node->name == "DATA-DEF" ) {
	datadefs.append(node);
      }
      node = dataschema->get_next_child(pos);
    }
    // now simplify data-defs
    GPList<XmlNode> newdatadefs;
    for(pos=datadefs; pos; ++pos) {
      GP<XmlNode> node = datadefs[pos];
      GPosition structrefpos = node->attrvalues.contains("STRUCTREF");
      GPosition namepos = node->attrvalues.contains("NAME");
      if( namepos && structrefpos) {
	GString fullstructrefname = node->attrvalues[structrefpos];
	// delete structref from datadef
	node->attrvalues.del(structrefpos);
	int poundpos = fullstructrefname.search('#');
	if( poundpos < 0 )
	  continue;
	GString structrefname = fullstructrefname.substr(poundpos+1, -1);
	GString name = node->attrvalues[namepos];
	for(GPosition structpos = datastructs; structpos; ++structpos) {
	  if( patternmatch(structrefname+".*", datastructs.key(structpos) ) ) {
	    GP<XmlNode> newdef = datastructs[structpos]->get_copy(dataschema);
	    newdef->name = "DATA-DEF";
	    newdef->attrvalues["NAME"] =
	      PrepareNewDataDefName(name, structrefname,
				    datastructs.key(structpos));
	    newdatadefs.append(newdef);
	  }
	}
      }
    }
    if( newdatadefs.size() ) {
      dataschema->add_children(newdatadefs);
      newdatadefs.empty();
      datastructs.empty();
      datadefs.empty();
      return SimplifyDataDefs(dataschema);
    }
    else {
      dataschema->delete_named_children("DATA-STRUCT");
      return 0;
    }
  }
  return 0;
}

//  simplyfy policy tree
bool P3PEvaluator::isDataDefHasCategories(GP<XmlNode> &datadef) {
  if( datadef && datadef->name == "DATA-DEF" ) {
    GPosition pos;
    GP<XmlNode> cat = datadef->get_first_child(pos);
    while(cat && cat->name != "CATEGORIES")
      cat = datadef->get_next_child(pos);

    if( cat )
      return true;
    else
      return false;
  }
  return false;
}

GP<XmlNode> P3PEvaluator::FindDataDef(GP<XmlNode> scheme, GString name ) {
  if( !scheme )
    return NULL;

  GPosition ddpos;
  GP<XmlNode> datadef = scheme->get_first_child(ddpos);
  while(datadef) {
    if( datadef->name == "DATA-DEF" ) {
      if( GPosition pos = datadef->attrvalues.contains("NAME") ) {
	if( datadef->attrvalues[pos] == name )
	  return datadef;
      }
    }
    datadef = scheme->get_next_child(ddpos);
  }
  // if that perticular name is not found, then try for superset
  GP<XmlNode> superdef;
  /*
    datadef = scheme->get_first_child(ddpos);
    while(datadef) {
    if( datadef->name == "DATA-DEF" ) {
    if( GPosition pos = datadef->attrvalues.contains("NAME") ) {
    if( patternmatch(name+".*", datadef->attrvalues[pos]) ) {
    if( superdef )
    superdef->add_children(datadef);
    else
    superdef = datadef->get_copy(NULL);
    }
    }
    }
    datadef = scheme->get_next_child(ddpos);
    }
    */

  return NULL;
}

bool P3PEvaluator::FindAttachSupersetDataDef(GP<XmlNode> nodeparent,
					     GP<XmlNode> node,
					     GP<XmlNode> scheme,
					     GString name, GString base) {
  if( !nodeparent || !node || !scheme)
    return false;
  bool flag = false;

  GPosition ddpos;
  GP<XmlNode> datadef = scheme->get_first_child(ddpos);
  while(datadef) {
    if( datadef->name == "DATA-DEF" ) {
      if( GPosition pos = datadef->attrvalues.contains("NAME") ) {
	if( patternmatch(name+".*", datadef->attrvalues[pos]) && !patternmatch(name+".*.*", datadef->attrvalues[pos])) {
	  if( !flag ) {
	    if( isDataDefHasCategories(datadef) ) {
	      node->delete_named_children("CATEGORIES");
	      node->attrvalues["REF"] = base + datadef->attrvalues[pos];
	      // node->attrvalues["##REF##"] = base + datadef->attrvalues[pos];
	      if( GPosition sdescpos = datadef->attrvalues.contains("SHORT-DESCRIPTION") )
		node->attrvalues["SHORT-DESCRIPTION"] = datadef->attrvalues[sdescpos];
	    }
	    node->add_children(datadef);
	  }
	  else {
	    GP<XmlNode> temp = new XmlNode(nodeparent);
	    temp->name = "DATA";
	    temp->attrvalues["REF"] = base + datadef->attrvalues[pos];
	    //temp->attrvalues["##REF##"] = base + datadef->attrvalues[pos];
	    if( GPosition sdescpos = datadef->attrvalues.contains("SHORT-DESCRIPTION") )
	      temp->attrvalues["SHORT-DESCRIPTION"] = datadef->attrvalues[sdescpos];
	    temp->add_children(datadef);
	    nodeparent->add_child(temp);
	  }
	  flag = true;
						
	}
      }
    }
    datadef = scheme->get_next_child(ddpos);
  }

	
  return flag;

}

GP<XmlNode> P3PEvaluator::FindDataSchemaAndSimplify(GP<XmlNode> p3ptree) {

  if( !p3ptree )
    return NULL;
  if( p3ptree->name == "DATASCHEMA") {
    SimplifyDataDefs(p3ptree);
    return p3ptree;
  }
  else if( p3ptree->name == "POLICY" || p3ptree->name == "POLICIES") {
    GPosition spos;
    GP<XmlNode> schema = p3ptree->get_first_child(spos);
    while(schema) {
      if( schema->name == "DATASCHEMA" ) {
	SimplifyDataDefs(schema);
	return schema;
      }
      else if( schema->name == "POLICY" ) {
	GPosition inspos;
	GP<XmlNode> inschema = schema->get_first_child(inspos);
	while(inschema) {
	  if( inschema->name == "DATASCHEMA" ) {
	    SimplifyDataDefs(inschema);
	    return inschema;
	  }
	  inschema = schema->get_next_child(inspos);
	}
      }
      schema = p3ptree->get_next_child(spos);
    }
  }
  return NULL;
}


void P3PEvaluator::AttachDataDef(GP<XmlNode> node, GP<XmlNode> datadef) {
  if( !node )
    return;

  if( GPosition sdescpos = datadef->attrvalues.contains("SHORT-DESCRIPTION") )
    node->attrvalues["SHORT-DESCRIPTION"] = datadef->attrvalues[sdescpos];

  if( isDataDefHasCategories(datadef) ) {
    node->delete_named_children("CATEGORIES");
  }
  else {
    node->attrvalues[VARIABLE_CATEGORY] = "YES";
  }

  node->add_children(datadef);
}

//simplifies and returns correct named policy
GP<XmlNode> P3PEvaluator::SimplifyPolicyTree(GP<XmlNode> p3ptree,
					     const GURL &policyurl,
					     GString policyname) {
  P3PDataSchema::Initialize();
  GP<XmlNode> basedataschema = P3PDataSchema::basedataschema;
  if(!basedataschema)
    return p3ptree;
  GMap<GURL, GP<XmlNode> > urldataschemamap;
  GP<XmlNode> inlineschema = FindDataSchemaAndSimplify(p3ptree);
  urldataschemamap[policyurl] = inlineschema;
  p3ptree = FindNamedPolicy(p3ptree, policyname);
  if( p3ptree && p3ptree->name == "POLICY" ) {
    GPosition spos;
    GP<XmlNode> statement = p3ptree->get_first_child(spos);
    while(statement) {
      if(statement->name == "STATEMENT" ) {
	GPosition dgpos;
	GP<XmlNode> datagroup = statement->get_first_child(dgpos);
	while( datagroup ) {
	  if( datagroup->name == "DATA-GROUP" ) {
	    // get the base name if there
	    GString base;
	    bool embeddedschema = false;
	    if( GPosition basepos = datagroup->attrvalues.contains("BASE") ) {
	      base = datagroup->attrvalues[basepos];
	      if( base == "" )
		embeddedschema = true;
	    }
	    // to coolect any new nodes because of #thirdparty kind of things
	    GP<XmlNode> tempnodes = new XmlNode(NULL);
	    GPosition datapos;
	    GP<XmlNode> data = datagroup->get_first_child(datapos);
	    while(data) {
	      if( data->name == "DATA" ) {
		// simplify here
		if( GPosition refpos = data->attrvalues.contains("REF") ) {
		  GString reference = base + data->attrvalues[refpos];
		  int poundpos = reference.search('#');
		  // if starting is '#' then it is refering basedatascema
		  if( !embeddedschema && poundpos == 0 ) {
		    GString refname = reference.substr(1,-1);
		    // find out whether it is fixed category if so delete
		    // any categories and add fixed categories
		    GP<XmlNode> datadef = FindDataDef(basedataschema, refname);
		    if( datadef ) {
		      AttachDataDef(data, datadef);
		    }
		    else if(!FindAttachSupersetDataDef(tempnodes, data,
						       basedataschema,
						       refname, "#" ))
		      THROW(GString("Data-def not found in the base data schema: ")+reference);
		  }
		  else if( embeddedschema && poundpos == 0) {
		    // process with the inline schema
		    GString refname = reference.substr(1,-1);
		    // find out whether it is fixed category if so delete
		    // any categories and add fixed categories
		    GP<XmlNode> datadef = FindDataDef(inlineschema, refname );
		    if( datadef ) {
		      AttachDataDef(data, datadef);
		    }
		    else if(!FindAttachSupersetDataDef(tempnodes, data,
						       basedataschema, refname,
						       GString(policyurl)+"#"))
		      THROW(GString("Data-def not found in the inpolicy data schema: ")+reference);
		  }
		  else {   // poundpos > 0
		    TRY {
		      GURL refurl = policyurl.add_relative(reference.substr(0, poundpos));
		      GString refname = reference.substr(poundpos+1, -1);
		      if( GPosition pos = urldataschemamap.contains(refurl) ) {
			GP<XmlNode> schemanode = urldataschemamap[pos];
			if( schemanode ) {
			  // find out whether it is fixed category if so
			  // delete any categories and add fixed categories
			  GP<XmlNode> datadef = FindDataDef(schemanode,
							    refname );
			  if( datadef ) {
			    AttachDataDef(data, datadef);
			  }
			  else if(!FindAttachSupersetDataDef(tempnodes, data,
							     schemanode,
							     refname,
							     GString(refurl)+
							     "#"))
			    THROW(GString("Data-def not found in the inpolicy data schema: ")+reference);
			}
			else
			  THROW(GString("Schema not found for policy : "+refurl));
		      }
		      else {
			// download the policy and find it
			char *policydata=g_policy_handler.GetURL(refurl);
			if( policydata ) {
			  GP<XmlNode> root = SaxXmlParse(policydata,
							 strlen(policydata),
							 true);
			  delete policydata;
			  policydata = 0;
			  GP<XmlNode> schemanode =
			    FindDataSchemaAndSimplify(root);
			  if( schemanode ) {
			    urldataschemamap[refurl] = schemanode;
			    GP<XmlNode> datadef =
			      FindDataDef(schemanode, refname );
			    if(datadef) {
			      AttachDataDef(data, datadef);
			    }
			    else if(!FindAttachSupersetDataDef(tempnodes,
							       data,
							       schemanode,
							       refname,
							       GString(refurl)
							       +"#" ))
			      THROW(GString("Data-def not found in the inpolicy data schema: ")+reference);
			  }
			  else
			    THROW(GString("Schema not found in policy : ")
				  +refurl);
			}
			else
			  THROW(GString("Can not download the the policy url:")
				+refurl);
		      }
		    }
		    CATCH(ex) {
		      THROW(ex.get_cause()+GString("Exception while processing data ref item :")+reference);
		    }
		    ENDCATCH;
		  }
		}
	      }
	      data = datagroup->get_next_child(datapos);
	    }
	    {
	      // add newly created DATA elements to DATA-GROUP
	      GPosition tpos;
	      GP<XmlNode> node = tempnodes->get_first_child(tpos);
	      while( node ) {
		datagroup->add_child(node);
		node = tempnodes->get_next_child(tpos);
	      }
	    }
	  }
	  datagroup = statement->get_next_child(dgpos);
	}
      }
      statement = p3ptree->get_next_child(spos);
    }
  }
  return p3ptree;
}

bool P3PEvaluator::IsExtensionRequired(GP<XmlNode> p3p_tree) {

  if( !p3p_tree )
    return false;
  if( p3p_tree->name == "EXTENSION" ) {
    if( GPosition optpos = p3p_tree->attrvalues.contains("OPTIONAL") ) {
      if( p3p_tree->attrvalues[optpos].upcase() == "NO" )
	return true;
    }
  }
  else {
    GPosition pos;
    GP<XmlNode> node = p3p_tree->get_first_child(pos);
    while(node) {
      if( IsExtensionRequired(node) )
	return true;
      node = p3p_tree->get_next_child(pos);
    }
  }
  return false;
}

bool P3PEvaluator::CheckPolicyName(GP<XmlNode> policy,
				   const GString &policyname) {

  if( policy->name == "POLICY" ) {
    if( !policyname.length() )
      return true;
    if( GPosition namepos = policy->attrvalues.contains("NAME") ) {
      if( policy->attrvalues[namepos] == policyname )
	return true;
    }
  }
  return false;
}

GP<XmlNode> P3PEvaluator::FindNamedPolicy(GP<XmlNode> policiesnode,
					  const GString &policyname) {
  if( !policiesnode )
    return NULL;
  if( policiesnode->name == "POLICY" ) {
    if( CheckPolicyName(policiesnode, policyname) )
      return policiesnode;
    else
      NULL;
  }
  if( policiesnode->name == "META" ) {
    // it is a policy reference file with policy may be in it
    GPosition pos;
    GP<XmlNode> policies = policiesnode->get_first_child(pos);
    while(policies) {
      if( policies->name == "POLICIES" ) {
	policiesnode = policies;
	break;
      }
      policies = policiesnode->get_next_child(pos);
    }
  }
  if( policiesnode && policiesnode->name == "POLICIES") {
    GPosition pos;
    GP<XmlNode> policy = policiesnode->get_first_child(pos);
    while( policy ) {
      if( policy->name == "POLICY" ) {
	if( !policyname.length() )
	  return policy;

	if( GPosition namepos = policy->attrvalues.contains("NAME") ) {
	  if( policy->attrvalues[namepos] == policyname )
	    return policy;
	}
      }
      policy = policiesnode->get_next_child(pos);
    }
  }
  return NULL;
}

int P3PEvaluator::P3PEvalutor_Match(GP<XmlNode> appel_tree,
				    GP<XmlNode> p3p_tree,
				    const GURL &policyurl,
				    const GString &policyname,
				    Result &result) {

  if( !appel_tree || !p3p_tree )
    return A_NULL_POINTER_ERROR;
  p3p_tree = SimplifyPolicyTree(p3p_tree, policyurl, policyname);
  if( !p3p_tree )
    return A_P3P_PARSING_ERROR;
  if( IsPolicyMissingRequiredFields(p3p_tree) )
    THROW("Policy missing required fields");
  if( IsExtensionRequired(p3p_tree) )
    THROW("Extension required, which is not supported");
  /* add p3p defaults */
  AppelMatcher matcher;
  matcher.match(appel_tree, p3p_tree, result);
  // get the summary of policy
  get_summary(p3p_tree, result);
  return A_OK;
}


int P3PEvaluator::P3PEvaluator_Match(const GString &appel_filename,
				     const GString &p3p_filename,
				     const GURL &policyurl,
				     const GString &policyname,
				     Result &result) {
  GP<XmlNode> p3p_tree, appel_tree;
  int read_appel = 0;

  appel_tree = SaxXmlParse(appel_filename, true);
  if( !appel_tree )
    return A_APPEL_PARSING_ERROR;
  p3p_tree = SaxXmlParse(p3p_filename, true);
  if( !p3p_tree )
    return A_P3P_PARSING_ERROR;
  return P3PEvalutor_Match(appel_tree,p3p_tree, policyurl, policyname, result);
}

int P3PEvaluator::P3PEvaluator_Match_Memory(const char *AppelBuffer,
					    int appelLength, 
					    const char *P3PBuffer,
					    int P3PLength, 
					    const GURL &policyurl,
					    const GString &policyname,
					    Result &result) {
  GP<XmlNode> p3p_tree, appel_tree;
  int read_appel = 0;

  appel_tree = SaxXmlParse(AppelBuffer, appelLength, true);
  if( !appel_tree )
    return A_APPEL_PARSING_ERROR;
  p3p_tree = SaxXmlParse(P3PBuffer, P3PLength, true);
  if( !p3p_tree )
    return A_P3P_PARSING_ERROR;
  return P3PEvalutor_Match(appel_tree,p3p_tree, policyurl, policyname, result);
}

bool P3PEvaluator::IsEntityMissingRequiredFields(GP<XmlNode> entitytree) {
  if( entitytree->name == "ENTITY" ) {
    GPosition pos;
    GP<XmlNode> node = entitytree->get_first_child(pos);
    const char *entityentries[] = {"DATA-GROUP"};
    int i=0;
    while(node && i<sizeof(entityentries)/sizeof(char*)) {
      if( node->name == entityentries[i] ) {
	i++;
      }
      node = entitytree->get_next_child(pos);
    }
    if( i == sizeof(entityentries)/sizeof(char*) )
      return false;
  }
  return true;
}


bool P3PEvaluator::IsStatementMissingRequiredFields(GP<XmlNode> statementtree){
  
  if( statementtree->name == "STATEMENT" ) {
    GPosition pos;
    const char *statemententries1[] = {"PURPOSE", "RECIPIENT", "RETENTION"};
    const char *statemententries2[] = {"NON-IDENTIFIABLE"};
    GP<XmlNode> node = statementtree->get_first_child(pos);
    int i=0;
    while(node && i<sizeof(statemententries1)/sizeof(char*)) {
      if( node->name == statemententries1[i] ) {
	i++;
      }
      node = statementtree->get_next_child(pos);
    }
    if( i == sizeof(statemententries1)/sizeof(char*) )
      return false;
    node = statementtree->get_first_child(pos);
    i=0;
    while(node && i<sizeof(statemententries2)/sizeof(char*)) {
      if( node->name == statemententries2[i] ) {
	i++;
      }
      node = statementtree->get_next_child(pos);
    }
    if( i == sizeof(statemententries2)/sizeof(char*) )
      return false;
  }
  return true;
}


bool P3PEvaluator::IsAccessMissingRequiredFields(GP<XmlNode> accesstree) {

  if( accesstree->name == "ACCESS" ) {
    GPosition pos;
    GP<XmlNode> node = accesstree->get_first_child(pos);
    const char *accessentries[] = {"NONIDENT", "ALL", "CONTACT-AND-OTHER",
				   "IDENT-CONTACT", "OTHER-CONTACT", "NONE"};
    int i;

    while(node) {
      for(i=0; i<sizeof(accessentries)/sizeof(char*); i++) {
	if( node->name == accessentries[i] )
	  break;
      }
      if( i < sizeof(accessentries)/sizeof(char*) )
	return false;
      node = accesstree->get_next_child(pos);
    }
  }
  return true;
}


bool P3PEvaluator::IsPolicyMissingRequiredFields(GP<XmlNode> policytree) {

  if( policytree->name == "POLICY" ) {
    GPosition pos;
    GP<XmlNode> node = policytree->get_first_child(pos);
    const char *policyentries[] = {"ENTITY", "ACCESS", "STATEMENT"};
    int i=0;
    while(node && i<sizeof(policyentries)/sizeof(char*)) {
      if( node->name == policyentries[i] ) {
	if( node->name == "ENTITY" && IsEntityMissingRequiredFields(node))
	  break;
	else
	  if( node->name == "ACCESS" && IsAccessMissingRequiredFields(node))
	    break;
	  else
	    if(node->name == "STATEMENT" &&
	       IsStatementMissingRequiredFields(node))
	      break;
	i++;
      }
      node = policytree->get_next_child(pos);
    }
    if( i == sizeof(policyentries)/sizeof(char*) )
      return false;
  }
  return true;
}

