#include "PolicySummary.h"
#include "Globals.h"
#include "SaxXmlParse.h"


bool P3PDataSchema::initialized = false;
GCriticalSection P3PDataSchema::class_lock;
GP<XmlNode> P3PDataSchema::basedataschema;

int P3PDataSchema::Initialize() {
  GCriticalSectionLock lock(&class_lock);
  if( initialized )
    return 0;

  basedataschema = SaxXmlParse(BASE_SCHEMA_FILE, true);
  if( basedataschema )
    P3PEvaluator::SimplifyDataDefs(basedataschema);

  initialized = true;

  return 1;
}

GString GetLongDescription(GP<XmlNode> node) {
  if( node && node->name == "DATA") {
    GPosition pos;
    GP<XmlNode> child = node->get_first_child(pos);
    while(child) {
      if( child->name == "LONG-DESCRIPTION" ) {
	return child->content;
      }
      child = node->get_next_child(pos);
    }
  }
  return "";
}



void GetAccessData(GP<XmlNode> &access, Result &result) {
  if( access && access->name == "ACCESS" ) {
    GPosition pos;
    GP<XmlNode> node = access->get_first_child(pos);
    if( node ) {
      if( node->name == "NONIDENT" ) {
	result.Access = "This site does not keep any information identified with you in its records";
      }
      else if( node->name == "ALL" ) {
	result.Access = "This site allows you to access all of the information identified with you in its records";
      }
      else if( node->name == "CONTACT-AND-OTHER" ) {
	result.Access = "This site allows you to access your contact information and certain other information about you from its records";
      }
      else if( node->name == "IDENT-CONTACT" ) {
	result.Access = "This site allows you to access your contact information from its records";
      }
      else if( node->name == "OTHER-IDENT" ) {
	result.Access = "This site allows you to access some of the information about you from its records";
      }
      else if( node->name == "NONE" ) {
	result.Access = "This site does not permit you to find out what information "
	  "about you is stored in their records";
      }
    }
  }
}


void GetEntityData(GP<XmlNode> &entity, Result &result) {
  if( entity && entity->name == "ENTITY" ) {
    GPosition pos;
    GP<XmlNode> data_group = entity->get_first_child(pos);
    if( data_group && data_group->name == "DATA-GROUP" ) {
      GPosition datapos;
      GP<XmlNode> data = data_group->get_first_child(datapos);
      while(data) {
	if(data->name == "DATA")
	  if( GPosition refpos = data->attrvalues.contains("REF") ) {
	    GString refvalue = data->attrvalues[refpos];

	    if( refvalue == "#business.name" ) {
	      result.BusinessName = data->content;
	    }
	    else if( refvalue == "#business.contact-info.postal.name" ) {
	      result.BusinessPostalName = data->content;
	    }
	    else if( refvalue == "#business.contact-info.postal.street" ) {
	      result.BusinessPostalStreet = data->content;
	    }
	    else if( refvalue == "#business.contact-info.postal.city" ) {
	      result.BusinessPostalCity = data->content;
	    }
	    else if( refvalue == "#business.contact-info.postal.stateprov" ) {
	      result.BusinessPostalState = data->content;
	    }
	    else if( refvalue == "#business.contact-info.postal.postalcode" ) {
	      result.BusinessPostalCode = data->content;
	    }
	    else if( refvalue == "#business.contact-info.postal.country" ) {
	      result.BusinessPostalCountry = data->content;
	    }
	    else if( refvalue == "#business.contact-info.telecom.telephone.intcode" ) {
	      result.BusinessTelephoneIntCode = data->content;
	    }
	    else if( refvalue == "#business.contact-info.telecom.telephone.loccode" ) {
	      result.BusinessTelephoneLocCode = data->content;
	    }
	    else if( refvalue == "#business.contact-info.telecom.telephone.number" ) {
	      result.BusinessTelephoneNumber = data->content;
	    }
	    else if( refvalue == "#business.contact-info.telecom.telephone.ext" ) {
	      result.BusinessTelephoneExt = data->content;
	    }
	    else if( refvalue == "#business.contact-info.telecom.telephone.comment" ) {
	      result.BusinessTelephoneComment = data->content;
	    }
	    else if( refvalue == "#business.contact-info.telecom.fax.intcode" ) {
	      result.BusinessFaxIntCode = data->content;
	    }
	    else if( refvalue == "#business.contact-info.telecom.fax.loccode" ) {
	      result.BusinessFaxLocCode = data->content;
	    }
	    else if( refvalue == "#business.contact-info.telecom.fax.number" ) {
	      result.BusinessFaxNumber = data->content;
	    }
	    else if( refvalue == "#business.contact-info.telecom.fax.ext" ) {
	      result.BusinessFaxExt = data->content;
	    }
	    else if( refvalue == "#business.contact-info.telecom.fax.comment" ) {
	      result.BusinessFaxComment = data->content;
	    }
	    else if( refvalue == "#business.contact-info.online.email" ) {
	      result.BusinessOnlineEmail = data->content;
	    }
	    else if( refvalue == "#business.contact-info.online.uri" ) {
	      result.BusinessOnlineURI = data->content;
	    }
	  }
	data = data_group->get_next_child(datapos);
      }
    }
  }
}



void GetExtensionData(GP<XmlNode> &extension, GP<StatementInfo> &statement_object) {
  if( extension && extension->name == "EXTENSION" ) {
    GPosition pos;
    GP<XmlNode> group_info = extension->get_first_child(pos);
    while(group_info) {
      if( group_info->name == "GROUP-INFO" ) {
	if( GPosition namepos = group_info->attrvalues.contains("NAME") ) {
	  statement_object->ExtensionName = group_info->attrvalues[namepos];
	  return;
	}
      }
      group_info = extension->get_next_child(pos);
    }
  }
}


void GetPurposeData(GP<XmlNode> &purpose, GP<StatementInfo> &statement_object) {
  if( purpose && purpose->name == "PURPOSE" ) {
    int other_purpose_count=1;

    GPosition pos;
    GP<XmlNode> node = purpose->get_first_child(pos);
    while(node) {
      GString purp;
      GString key = node->name;
			
      if( node->name == "CURRENT" ) {
	purp = "To complete the activity "
	  "for which the data was provided";
      }
      else if( node->name == "ADMIN" ) {
	purp = "To do web site and system administration";
      }
      else if( node->name == "DEVELOP" ) {
	purp = "To do research and development without creating a record identified with you";
      }
      else if( node->name == "CUSTOMIZATION" ) {
	purp = "To provide the custom services you requested";
      }
      else if( node->name == "TAILORING" ) {
	purp = "To customize the site for your current visit";
      }
      else if( node->name == "PSEUDO-ANALYSIS" ) {
	purp = "To do research and analysis in which your records may be used without identifying you";
      }
      else if( node->name == "PSEUDO-DECISION" ) {
	purp = "To make decisions - for example, about what content or ads you see at the web site - without identifying you";
      }
      else if( node->name == "INDIVIDUAL-ANALYSIS" ) {
	purp = "To do research and analysis in which your records may be used";
      }
      else if( node->name == "INDIVIDUAL-DECISION" ) {
	purp = "To make decisions - for example, about what content or ads you see at the web site";
      }
      else if( node->name == "CONTACT" ) {
	purp = "To contact you through means other than telephone "
	  "(email, postal mail, etc.) to interest you "
	  "in other services or products";
      }
      else if( node->name == "HISTORICAL" ) {
	purp = "To preserve social history";
      }
      else if( node->name == "TELEMARKETING" ) {
	purp = "To contact you by telephone to interest you "
	  "in other services or products";
      }
      else if( node->name == "OTHER-PURPOSE" ) {
	purp = "Other unknown uses: "+ node->content;
	key += (GString)other_purpose_count;
	other_purpose_count++;
      }
			
      if( GPosition requiredpos = node->attrvalues.contains("REQUIRED") ) {
	GString required = node->attrvalues[requiredpos].upcase();

	if( required == "OPT-IN" )
	  purp+= " --  only if you request this";
	else if( required == "OPT-OUT" )
	  purp+= " --  unless you #opt-out#";
      }

      //result.Purposes[node->name] = purp;
      if( purp.length() )
	statement_object->Purposes[key] = purp;

      node = purpose->get_next_child(pos);
    }
  }
}


void GetRecipientData(GP<XmlNode> &recipient, GP<StatementInfo> &statement_object) {
  if( recipient && recipient->name == "RECIPIENT" ) {
    GPosition pos;
    GP<XmlNode> node = recipient->get_first_child(pos);
    while(node) {
      GString recp;

      if( node->name == "OURS" ) {
	recp = "This web site and the companies that help the site provide services to you (such companies must use your information only on behalf of this web site for the purposes stated in this policy)";
      }
      else if( node->name == "DELIVERY" ) {
	recp = "Delivery companies who may also use "
	  "your information for other purposes";
      }
      else if( node->name == "SAME" ) {
	recp = "Other companies with similar privacy policies";
      }
      else if( node->name == "OTHER-RECIPIENT" ) {
	recp = "Other companies with different privacy policies";
      }
      else if( node->name == "UNRELATED" ) {
	recp = "Other companies";
      }
      else if( node->name == "PUBLIC" ) {
	recp = "People who retrieve your information from public "
	  "bulletin boards, chat rooms, directories, etc.";
      }

      if( GPosition requiredpos = node->attrvalues.contains("REQUIRED") ) {
	GString required = node->attrvalues[requiredpos].upcase();

	if( required == "OPT-IN" )
	  recp += " --  only if you request this";
	else if( required == "OPT-OUT" )
	  recp += " --  unless you #opt-out#";
      }
      if( recp.length() )
	statement_object->Recipients[node->name] = recp;
      node = recipient->get_next_child(pos);
    }
  }
}


GMap<GString, GString> GetCatogoryInfo(GP<XmlNode> catnode, GString optional) {
  GMap<GString, GString> catmap;
  if( catnode->name == "CATEGORIES" ) {
    int other_category_count=1;
    GPosition categoriespos;
    GP<XmlNode> categories = catnode->get_first_child(categoriespos);
		
    while(categories) {
      GString key = categories->name;
      GString cat;
      if( categories->name == "PHYSICAL" ) {
	cat = "Name, address, phone number, or other "
	  "contact information";
      }
      else if( categories->name == "ONLINE" ) {
	cat = " Email address or other "
	  "online contact information";
      }
      else if( categories->name == "UNIQUEID" ) {
	cat = " Website login IDs and other identifiers "
	  "(excluding government IDs and financial account numbers)";
      }
      else if( categories->name == "PURCHASE" ) {
	cat = "Information about your purchases, "
	  "including information about payment methods";
      }
      else if( categories->name == "FINANCIAL" ) {
	cat = "Information about your "
	  "finances such as account balances, payment "
	  "or overdraft history, loans, etc.";
      }
      else if( categories->name == "COMPUTER" ) {
	cat = "Information about the kind of "
	  "computer you are using, the software you are "
	  "running on it, and its Internet address";
      }
      else if( categories->name == "NAVIGATION" ) {
	cat = "Information about which pages you "
	  "visited on this web site and how long you "
	  "stayed at each page ";
      }
      else if( categories->name == "INTERACTIVE" ) {
	cat = "Information about activities you "
	  "engaged in at this web site, such as searches "
	  "you performed or logs of account activity";
      }
      else if( categories->name == "DEMOGRAPHIC" ) {
	cat = "Your demographic and socioeconomic data, "
	  "such as gender, age, and income";
      }
      else if( categories->name == "CONTENT" ) {
	cat = "The text of email you send to this "
	  "site, bulletin board postings, chat room "
	  "communications, etc. ";
      }
      else if( categories->name == "STATE" ) {
	cat = "Cookies and related mechanisms ";
      }
      else if( categories->name == "POLITICAL" ) {
	cat = "Information about your affiliation "
	  "with groups such as religious organizations, "
	  "trade unions, professional associations, "
	  "political parties, etc.";
      }
      else if( categories->name == "HEALTH" ) {
	cat = "Information about your health, including information about purchases of healthcare products and services";
      }
      else if( categories->name == "PREFERENCE" ) {
	cat = "Information about your tastes "
	  "or preferences";
      }
      else if( categories->name == "LOCATION" ) {
	cat = "Information about your current physical location";
      }
      else if( categories->name == "GOVERNMENT" ) {
	cat = "Government-issued identifiers "
	  "such as social security numbers";
      }
      else if( categories->name == "OTHER-CATEGORY" ) {
	cat = "Other unknown types of data: "+ categories->content;
	key += (GString)other_category_count;
	other_category_count++;
      }

      if( optional == "YES" )
	cat += " (optional)";

      if( cat.length() )
	catmap[key] = cat;
      categories = catnode->get_next_child(categoriespos);
    }
  }
  return catmap;
}



void GetDataGroupData(GP<XmlNode> &data_group, GP<StatementInfo> &statement_object) {
  if( data_group && data_group->name == "DATA-GROUP" ) {

    GPosition pos;
    GP<XmlNode> data = data_group->get_first_child(pos);
    while(data) {
      if( data->name == "DATA" ) {
	GString cat;
	GString optional;

	if( GPosition optionalpos = data->attrvalues.contains("OPTIONAL") ) {
	  optional = data->attrvalues[optionalpos].upcase();
	}

	GString longdesc = GetLongDescription(data);
	GPosition sdescpos = data->attrvalues.contains("SHORT-DESCRIPTION");
	GPosition refpos = data->attrvalues.contains("REF");
	/*if( !refpos )
	  refpos = data->attrvalues.contains("##REF##");*/
	if( (longdesc.length() || sdescpos) && refpos) {
	  if( longdesc.length() )
	    cat = longdesc;
	  else
	    cat = data->attrvalues[sdescpos];

	  if( optional == "YES" )
	    cat += " (optional)";

	  DataInfo datainfo = cat;

	  if( data->has_children() && data->attrvalues.contains(VARIABLE_CATEGORY)) {
	    GPosition datapos;
	    GP<XmlNode> node = data->get_first_child(datapos);
	    while(node) {
	      if( node->name == "CATEGORIES" ) {
		datainfo.CatInfo = GetCatogoryInfo(node, optional);
								
	      }
	      node = data->get_next_child(datapos);
	    }
	  }
	  statement_object->InfoCollected[data->attrvalues[refpos]] = datainfo;

	}
	else {
	  //////////////////////////////////////////////////////////////
	  GPosition datapos;
	  GP<XmlNode> node = data->get_first_child(datapos);

	  while(node) {
	    if( node->name == "CATEGORIES" ) {
	      GMap<GString, GString> catmap = GetCatogoryInfo(node, optional);
	      for(GPosition catpos=catmap; catpos; ++catpos)
		statement_object->InfoCollected[catmap.key(catpos)] = catmap[catpos];

	      /*
		int other_category_count=1;
		GPosition categoriespos;
		GP<XmlNode> categories = node->get_first_child(categoriespos);
							
		while(categories) {
		GString key = categories->name;
		GString cat;
		if( categories->name == "PHYSICAL" ) {
		cat = "Name, address, phone number, or other "
		"contact information";
		}
		else if( categories->name == "ONLINE" ) {
		cat = " Email address or other "
		"online contact information";
		}
		else if( categories->name == "UNIQUEID" ) {
		cat = " Website login IDs and other identifiers "
		"(excluding government IDs and financial account numbers)";
		}
		else if( categories->name == "PURCHASE" ) {
		cat = "Information about your purchases, "
		"including information about payment methods";
		}
		else if( categories->name == "FINANCIAL" ) {
		cat = "Information about your "
		"finances such as account balances, payment "
		"or overdraft history, loans, etc.";
		}
		else if( categories->name == "COMPUTER" ) {
		cat = "Information about the kind of "
		"computer you are using, the software you are "
		"running on it, and its Internet address";
		}
		else if( categories->name == "NAVIGATION" ) {
		cat = "Information about which pages you "
		"visited on this web site and how long you "
		"stayed at each page ";
		}
		else if( categories->name == "INTERACTIVE" ) {
		cat = "Information about activities you "
		"engaged in at this web site, such as searches "
		"you performed or logs of account activity";
		}
		else if( categories->name == "DEMOGRAPHIC" ) {
		cat = "Your demographic and socioeconomic data, "
		"such as gender, age, and income";
		}
		else if( categories->name == "CONTENT" ) {
		cat = "The text of email you send to this "
		"site, bulletin board postings, chat room "
		"communications, etc. ";
		}
		else if( categories->name == "STATE" ) {
		cat = "Cookies and related mechanisms ";
		}
		else if( categories->name == "POLITICAL" ) {
		cat = "Information about your affiliation "
		"with groups such as religious organizations, "
		"trade unions, professional associations, "
		"political parties, etc.";
		}
		else if( categories->name == "HEALTH" ) {
		cat = "Information about your health, including information about purchases of healthcare products and services";
		}
		else if( categories->name == "PREFERENCE" ) {
		cat = "Information about your tastes "
		"or preferences";
		}
		else if( categories->name == "LOCATION" ) {
		cat = "Information about your current physical location";
		}
		else if( categories->name == "GOVERNMENT" ) {
		cat = "Government-issued identifiers "
		"such as social security numbers";
		}
		else if( categories->name == "OTHER-CATEGORY" ) {
		cat = "Other unknown types of data: "+ categories->content;
		key += (GString)other_category_count;
		other_category_count++;
		}

		if( optional == "YES" )
		cat += " (optional)";

		if( cat.length() )
		statement_object->InfoCollected[key] = cat;
		categories = node->get_next_child(categoriespos);
		}
		*/
	    }
	    node = data->get_next_child(datapos);
	  }
	}
					
      }
      data = data_group->get_next_child(pos);
    }
  }
}



void GetConsequenceData(GP<XmlNode> &consequence, GP<StatementInfo> &statement_object) {
  if( consequence && consequence->name == "CONSEQUENCE") {
    statement_object->Consequence = consequence->content;
  }

}
	

void GetStatementData(GP<XmlNode> &statement, Result &result) {
  if( statement && statement->name == "STATEMENT" ) {
    GP<StatementInfo> statement_object = new StatementInfo();

    GPosition pos;
    GP<XmlNode> node = statement->get_first_child(pos);
    while(node) {
      if(node->name == "EXTENSION") {
	GetExtensionData(node, statement_object);
      }
      else if(node->name == "PURPOSE") {
	GetPurposeData(node, statement_object);
      }
      else if(node->name == "RECIPIENT") {
	GetRecipientData(node, statement_object);
      }
      else if(node->name == "DATA-GROUP") {
	GetDataGroupData(node, statement_object);
      }
      else if(node->name == "CONSEQUENCE") {
	GetConsequenceData(node, statement_object);
      }

      node = statement->get_next_child(pos);
    }

    result.Statements.append(statement_object);
  }
}


void GetDisputesGroupData(GP<XmlNode> &disputes_group, Result &result) {
  if( disputes_group && disputes_group->name == "DISPUTES-GROUP" ) {
    GPosition pos;
    GP<XmlNode> disputes = disputes_group->get_first_child(pos);
    while(disputes) {
      if(disputes->name == "DISPUTES") {
	GP<DisputeInfo> dispute_object = new DisputeInfo();
	if( GPosition attrpos = disputes->attrvalues.contains("SERVICE") ) {
	  dispute_object->ServiceURL = disputes->attrvalues[attrpos];
	}
	if( GPosition attrpos = disputes->attrvalues.contains("SHORT-DESCRIPTION") ) {
	  dispute_object->ShortDescription = disputes->attrvalues[attrpos];
	}

	GPosition disputespos;
	GP<XmlNode> node = disputes->get_first_child(disputespos);
	while(node) {
	  if(node->name == "LONG-DESCRIPTION") {
	    dispute_object->LongDescription = node->content;
	  }
	  else if(node->name == "IMG") {
	    if( GPosition attrpos = node->attrvalues.contains("SRC") ) {
	      dispute_object->ImageURL = node->attrvalues[attrpos];
	    }
	    if( GPosition attrpos = node->attrvalues.contains("ALT") ) {
	      dispute_object->ImgAltText = node->attrvalues[attrpos];
	    }
	  }
	  node = disputes->get_next_child(disputespos);
	} //end while(node)

	result.Disputes.append(dispute_object);
      }
      disputes = disputes_group->get_next_child(pos);
    }
  }

}


void GetExpiryData(GP<XmlNode> &expiry, Result &result) {
  if( expiry && expiry->name == "EXPIRY") {
    if( GPosition attrpos = expiry->attrvalues.contains("DATE") ) {
      result.Expiry = expiry->attrvalues[attrpos];
    }
    else if( GPosition attrpos = expiry->attrvalues.contains("MAX-AGE") ) {
      result.Expiry = expiry->attrvalues[attrpos] + " seconds";
    }
  }

}
	

int get_summary(GP<XmlNode> &p3p_tree, Result &result) {
  if( p3p_tree && p3p_tree->name == "POLICY" ) {
    GPosition pos;
    result.Test = false;

    GP<XmlNode> node = p3p_tree->get_first_child(pos);
    while(node) {
      if( node->name == "ACCESS" )
	GetAccessData(node, result);
      else if( node->name == "ENTITY")
	GetEntityData(node, result);
      else if( node->name == "STATEMENT")
	GetStatementData(node, result);
      else if( node->name == "DISPUTES-GROUP")
	GetDisputesGroupData(node, result);
      else if( node->name == "EXPIRY")
	GetExpiryData(node, result);
      else if( node->name == "TEST")
	result.Test = true;

      node = p3p_tree->get_next_child(pos);
    }
    return 1;
  }
  return 0;
}
